import re
import sys
from unittest.mock import MagicMock, _Call, patch

import pytest
from sentence_transformers import SentenceTransformer

from fed_rag.base.retriever import BaseRetriever
from fed_rag.exceptions import MissingExtraError
from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
    HFSentenceTransformerRetriever,
    InvalidLoadType,
    LoadKwargs,
)


def test_hf_sentence_transformer_retriever_class() -> None:
    names_of_base_classes = [
        b.__name__ for b in HFSentenceTransformerRetriever.__mro__
    ]
    assert BaseRetriever.__name__ in names_of_base_classes


@patch.object(HFSentenceTransformerRetriever, "_load_model_from_hf")
def test_hf_sentence_transformer_retriever_class_init_delayed_load(
    mock_load_from_hf: MagicMock,
    dummy_sentence_transformer: SentenceTransformer,
) -> None:
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name", load_model_at_init=False
    )

    assert retriever.model_name == "fake_name"
    assert retriever._encoder is None
    assert retriever._query_encoder is None
    assert retriever._context_encoder is None

    # load model
    mock_load_from_hf.return_value = dummy_sentence_transformer
    retriever._load_model_from_hf(load_type="encoder")
    args, kwargs = mock_load_from_hf.call_args

    # assert
    mock_load_from_hf.assert_called_once()
    assert retriever.encoder == dummy_sentence_transformer
    assert args == ()
    assert kwargs == {"load_type": "encoder"}


@patch.object(HFSentenceTransformerRetriever, "_load_model_from_hf")
def test_hf_sentence_transformer_retriever_class_init_delayed_dual_encoder_load(
    mock_load_from_hf: MagicMock,
    dummy_sentence_transformer: SentenceTransformer,
) -> None:
    retriever = HFSentenceTransformerRetriever(
        query_model_name="query_fake_name",
        context_model_name="context_fake_name",
        load_model_at_init=False,
    )

    assert retriever.model_name is None
    assert retriever.query_model_name == "query_fake_name"
    assert retriever.context_model_name == "context_fake_name"
    assert retriever._encoder is None
    assert retriever._query_encoder is None
    assert retriever._context_encoder is None

    # load models
    mock_load_from_hf.return_value = dummy_sentence_transformer
    retriever._load_model_from_hf(load_type="query_encoder")
    retriever._load_model_from_hf(load_type="context_encoder")

    # assert
    calls = [
        _Call(((), {"load_type": "query_encoder"})),
        _Call(((), {"load_type": "context_encoder"})),
    ]
    mock_load_from_hf.assert_has_calls(calls)
    assert retriever.query_encoder == dummy_sentence_transformer
    assert retriever.context_encoder == dummy_sentence_transformer


@patch.object(HFSentenceTransformerRetriever, "_load_model_from_hf")
def test_hf_sentence_transformer_retriever_class_init_dual_encoder(
    mock_load_from_hf: MagicMock,
    dummy_sentence_transformer: SentenceTransformer,
) -> None:
    # arrange
    mock_load_from_hf.return_value = dummy_sentence_transformer

    # act
    retriever = HFSentenceTransformerRetriever(
        query_model_name="query_fake_name",
        context_model_name="context_fake_name",
    )

    # assert
    calls = [
        _Call(((), {"load_type": "query_encoder"})),
        _Call(((), {"load_type": "context_encoder"})),
    ]
    mock_load_from_hf.assert_has_calls(calls)
    assert retriever.encoder is None
    assert retriever.query_encoder == dummy_sentence_transformer
    assert retriever.context_encoder == dummy_sentence_transformer


@patch.object(HFSentenceTransformerRetriever, "_load_model_from_hf")
def test_hf_sentence_transformer_retriever_class_init_encoder(
    mock_load_from_hf: MagicMock,
    dummy_sentence_transformer: SentenceTransformer,
) -> None:
    # arrange
    mock_load_from_hf.return_value = dummy_sentence_transformer

    # act
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name",
    )

    # assert
    calls = [
        _Call(((), {"load_type": "encoder"})),
    ]
    mock_load_from_hf.assert_has_calls(calls)
    assert retriever.encoder == dummy_sentence_transformer
    assert retriever.query_encoder is None
    assert retriever.context_encoder is None


@patch(
    "fed_rag.retrievers.huggingface.hf_sentence_transformer.SentenceTransformer"
)
def test_load_model_from_hf_constructs_sentence_transformer_obj(
    mock_sentence_transformer: MagicMock,
) -> None:
    # arrange
    # act
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name", load_model_kwargs={"device": "cpu"}
    )

    # assert
    mock_sentence_transformer.assert_called_once_with(
        "fake_name", device="cpu"
    )
    assert retriever.query_encoder is None
    assert retriever.context_encoder is None


@patch(
    "fed_rag.retrievers.huggingface.hf_sentence_transformer.SentenceTransformer"
)
def test_load_model_from_hf_constructs_sentence_transformer_obj_dual(
    mock_sentence_transformer: MagicMock,
) -> None:
    # arrange
    # act
    retriever = HFSentenceTransformerRetriever(
        query_model_name="fake_query_model_name",
        context_model_name="fake_context_model_name",
        load_model_kwargs=LoadKwargs(
            query_encoder={"device": "cuda:0"},
            context_encoder={"device": "cpu"},
        ),
    )

    # assert
    mock_sentence_transformer.assert_has_calls(
        [
            _Call((("fake_query_model_name",), {"device": "cuda:0"})),
            _Call((("fake_context_model_name",), {"device": "cpu"})),
        ]
    )
    assert retriever.encoder is None


def test_load_model_from_hf_raises_invalid_type_error() -> None:
    # arrange
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name", load_model_at_init=False
    )

    # act/assert
    with pytest.raises(InvalidLoadType, match="Invalid `load_type` supplied."):
        retriever._load_model_from_hf(load_type="unsupported_type")


@patch.object(HFSentenceTransformerRetriever, "_load_model_from_hf")
def test_encode_query(
    mock_load_from_hf: MagicMock,
) -> None:
    # arrange
    mock_encoder = MagicMock()
    mock_encoder.encode.side_effect = iter([[1, 2, 3], [4, 5, 6]])
    mock_load_from_hf.return_value = mock_encoder
    retriever = HFSentenceTransformerRetriever(
        model_name="query_fake_name",
    )

    # act
    context_emb = retriever.encode_context("fake_context")
    query_emb = retriever.encode_query("fake_query")

    # assert
    assert context_emb == [1, 2, 3]
    assert query_emb == [4, 5, 6]


def test_huggingface_extra_missing() -> None:
    """Test extra is not installed."""

    modules = {"transformers": None, "sentence_transformers": None}
    module_to_import = "fed_rag.retrievers.huggingface.hf_sentence_transformer"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HFSentenceTransformerRetriever` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
                HFSentenceTransformerRetriever,
            )

            HFSentenceTransformerRetriever(
                model_name="query_fake_name",
            )

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module



import inspect

from fed_rag.base.knowledge_store import BaseKnowledgeStore


def test_base_abstract_attr() -> None:
    abstract_methods = BaseKnowledgeStore.__abstractmethods__

    assert inspect.isabstract(BaseKnowledgeStore)
    assert "load_node" in abstract_methods
    assert "load_nodes" in abstract_methods
    assert "retrieve" in abstract_methods
    assert "delete_node" in abstract_methods
    assert "clear" in abstract_methods
    assert "count" in abstract_methods



"""HuggingFace Data Collator For LM-Supervised Retriever Training"""

from typing import Any, Callable

import torch
from pydantic import Field, PrivateAttr

from fed_rag.base.data_collator import BaseDataCollator
from fed_rag.exceptions import MissingExtraError
from fed_rag.exceptions.core import FedRAGError
from fed_rag.types.rag_system import RAGSystem
from fed_rag.utils.huggingface import _validate_rag_system

try:
    from sentence_transformers.data_collator import (
        SentenceTransformerDataCollator,
    )

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

    # Create a dummy class with a different name to avoid the redefinition
    class _SentenceTransformerDataCollator:
        """Dummy placeholder when transformers is not available."""

        pass

    SentenceTransformerDataCollator = _SentenceTransformerDataCollator  # type: ignore


DEFAULT_PROMPT_TEMPLATE = """
You are a helpful assistant. Given the user's question, provide a succinct
and accurate response. If context is provided, use it in your answer if it helps
you to create the most accurate response.

<warning>
Only use the the provided context if its relevant to answer the question. Otherwise,
ignore it and use your parametric knowledge to answer the question.
</warning>

<question>
{query}
</question>

<context>
{context}
</context>

"""

DEFAULT_TARGET_TEMPLATE = """
<response>
{response}
</response>
"""


class DataCollatorForLSR(SentenceTransformerDataCollator, BaseDataCollator):
    """A HuggingFace DataCollator for LM-Supervised Retrieval."""

    prompt_template: str = Field(default=DEFAULT_PROMPT_TEMPLATE)
    target_template: str = Field(default=DEFAULT_TARGET_TEMPLATE)
    default_return_tensors: str = Field(default="pt")

    # Add these fields to make Pydantic aware of them
    tokenize_fn: Callable = Field(
        default_factory=lambda: (lambda *args, **kwargs: {})
    )
    valid_label_columns: list[str] = Field(
        default_factory=lambda: ["label", "score"]
    )
    _warned_columns: set = PrivateAttr(
        default_factory=set
    )  # exclude=True to match dataclass repr=False

    def __init__(
        self,
        rag_system: RAGSystem,
        prompt_template: str | None = None,
        target_template: str | None = None,
        default_return_tensors: str = "pt",
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        _validate_rag_system(rag_system)

        prompt_template = prompt_template or DEFAULT_PROMPT_TEMPLATE
        target_template = target_template or DEFAULT_TARGET_TEMPLATE

        # init pydantic base model
        BaseDataCollator.__init__(
            self,
            rag_system=rag_system,
            prompt_template=prompt_template,
            target_template=target_template,
            default_return_tensors=default_return_tensors,
            tokenize_fn=lambda *args, **kwargs: {},  # Pass this to Pydantic
            **kwargs,
        )

    def __call__(
        self, features: list[dict[str, Any]], return_tensors: str | None = None
    ) -> dict[str, Any]:
        """Use the features of the dataset in order to get the retrieval and lm-scores.


        Args:
            features (list[Any]): Should contain a 'query' and 'reponse' field.
            return_tensors (_type_, optional): supports right now only 'pt'

        Returns:
            dict[str, Any]: a dictionary of ~torch.Tensors with keys 'retrieval_scores'
                and 'lm_scores'
            Note that each ('query', 'response') pair generates one fine-tuning instance for LSR.
        """
        return_tensors = (
            return_tensors if return_tensors else self.default_return_tensors
        )
        if return_tensors != "pt":
            raise FedRAGError(f"Framework '{return_tensors}' not recognized!")

        # use rag system to get scores
        batch_retriever_scores = []
        batch_lm_scores = []
        for example in features:
            query = example.get("query")
            response = example.get("response")

            # retriever scores - this should participate in gradient computation
            source_nodes = self.rag_system.retrieve(query)
            retriever_scores = torch.tensor(
                [n.score for n in source_nodes], requires_grad=True
            )

            # lm supervised scores - we don't want these to participate in gradient computation
            lm_scores = []
            with torch.no_grad():
                for chunk in source_nodes:
                    prompt = self.prompt_template.format(
                        query=query,
                        context=chunk.node.get_content()["text_content"],
                    )
                    target = self.target_template.format(response=response)
                    lm_score = self.rag_system.generator.compute_target_sequence_proba(
                        prompt=prompt, target=target
                    )
                    lm_scores.append(lm_score)
                lm_scores = torch.stack(lm_scores, dim=0)

            # append to batch
            batch_retriever_scores.append(retriever_scores)
            batch_lm_scores.append(lm_scores)

        # create torch.Tensors
        retrieval_scores = torch.stack(batch_retriever_scores, dim=0)
        lm_scores = torch.stack(batch_lm_scores, dim=0)

        return {"retrieval_scores": retrieval_scores, "lm_scores": lm_scores}



import re
import sys
from unittest.mock import MagicMock, patch

import pytest
import torch
from peft import PeftModel
from torch.testing import assert_close
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    PreTrainedModel,
    PreTrainedTokenizer,
)

from fed_rag.base.generator import BaseGenerator
from fed_rag.base.tokenizer import EncodeResult
from fed_rag.exceptions import MissingExtraError
from fed_rag.generators.huggingface import HFPeftModelGenerator
from fed_rag.tokenizers.hf_pretrained_tokenizer import HFPretrainedTokenizer


def test_hf_peft_generator_class() -> None:
    names_of_base_classes = [b.__name__ for b in HFPeftModelGenerator.__mro__]
    assert BaseGenerator.__name__ in names_of_base_classes


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
@patch.object(HFPeftModelGenerator, "_load_model_from_hf")
def test_hf_peft_generator_class_init_delayed_load(
    mock_load_from_hf: MagicMock,
    mock_load_from_hf_tokenizer: MagicMock,
    dummy_peft_model_and_tokenizer: tuple[PeftModel, PreTrainedTokenizer],
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )

    assert generator.model_name == "fake_name"
    assert generator.base_model_name == "fake_base_name"
    assert generator._model is None
    assert generator.tokenizer is not None

    # load model
    mock_load_from_hf_tokenizer.return_value = dummy_peft_model_and_tokenizer[
        1
    ]
    mock_load_from_hf.return_value = dummy_peft_model_and_tokenizer[0]

    generator._load_model_from_hf()
    args, kwargs = mock_load_from_hf.call_args

    mock_load_from_hf.assert_called_once()
    assert generator.model == dummy_peft_model_and_tokenizer[0]
    assert isinstance(generator.tokenizer, HFPretrainedTokenizer)
    assert generator.tokenizer.unwrapped == dummy_peft_model_and_tokenizer[1]
    assert args == ()
    assert kwargs == {}


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
@patch.object(HFPeftModelGenerator, "_load_model_from_hf")
def test_hf_peft_generator_class_init(
    mock_load_from_hf: MagicMock,
    mock_load_from_hf_tokenizer: MagicMock,
    dummy_peft_model_and_tokenizer: tuple[PeftModel, PreTrainedTokenizer],
) -> None:
    # arrange
    mock_load_from_hf.return_value = dummy_peft_model_and_tokenizer[0]
    mock_load_from_hf_tokenizer.return_value = dummy_peft_model_and_tokenizer[
        1
    ]

    # act
    generator = HFPeftModelGenerator(
        model_name="fake_name", base_model_name="fake_base_name"
    )
    args, kwargs = mock_load_from_hf.call_args

    # assert
    mock_load_from_hf.assert_called_once()
    mock_load_from_hf_tokenizer.assert_called_once()
    assert generator.model_name == "fake_name"
    assert generator.base_model_name == "fake_base_name"
    assert generator.model == dummy_peft_model_and_tokenizer[0]
    assert isinstance(generator.tokenizer, HFPretrainedTokenizer)
    assert generator.tokenizer.unwrapped == dummy_peft_model_and_tokenizer[1]
    assert args == ()
    assert kwargs == {}


@patch.object(HFPretrainedTokenizer, "_load_model_from_hf")
@patch.object(HFPeftModelGenerator, "_load_model_from_hf")
def test_hf_pretrained_generator_class_init_no_load(
    mock_load_from_hf: MagicMock,
    mock_load_from_hf_tokenizer: MagicMock,
    dummy_peft_model_and_tokenizer: tuple[PeftModel, PreTrainedTokenizer],
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )

    mock_load_from_hf.assert_not_called()
    mock_load_from_hf_tokenizer.assert_not_called()
    assert generator.model_name == "fake_name"
    assert generator.base_model_name == "fake_base_name"
    assert generator._model is None
    assert generator._tokenizer is not None
    assert isinstance(generator.tokenizer, HFPretrainedTokenizer)

    # load model using setter
    model, tokenizer = dummy_peft_model_and_tokenizer
    generator.model = model
    generator.tokenizer.unwrapped = tokenizer

    assert generator.model == model
    assert generator.tokenizer.unwrapped == tokenizer


@patch(
    "fed_rag.generators.huggingface.hf_peft_model.prepare_model_for_kbit_training"
)
@patch.object(PeftModel, "from_pretrained")
@patch.object(AutoModelForCausalLM, "from_pretrained")
@patch.object(AutoTokenizer, "from_pretrained")
def test_hf_peft_load_model_from_hf(
    mock_tokenizer_from_pretrained: MagicMock,
    mock_auto_model_from_pretrained: MagicMock,
    mock_peft_model_from_pretrained: MagicMock,
    mock_prepare_model_for_kbit_training: MagicMock,
    dummy_pretrained_model_and_tokenizer: tuple[
        PreTrainedModel, PreTrainedTokenizer
    ],
    dummy_peft_model_and_tokenizer: tuple[PeftModel, PreTrainedTokenizer],
) -> None:
    # arrange
    model, tokenizer = dummy_peft_model_and_tokenizer
    base_model, _ = dummy_pretrained_model_and_tokenizer
    mock_auto_model_from_pretrained.return_value = base_model
    mock_prepare_model_for_kbit_training.return_value = base_model
    mock_peft_model_from_pretrained.return_value = model
    mock_tokenizer_from_pretrained.return_value = tokenizer

    # act
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_base_model_kwargs={
            "device_map": "cpu",
            "quantization_config": "fake config",
        },
        load_model_kwargs={"fake_param": "fake_value"},
    )

    # assert
    assert generator.model_name == "fake_name"
    assert generator.base_model_name == "fake_base_name"
    mock_tokenizer_from_pretrained.assert_called_once_with("fake_base_name")
    mock_auto_model_from_pretrained.assert_called_once_with(
        "fake_base_name", device_map="cpu", quantization_config="fake config"
    )
    mock_prepare_model_for_kbit_training.assert_called_once_with(base_model)
    mock_peft_model_from_pretrained.assert_called_once_with(
        base_model, "fake_name", fake_param="fake_value"
    )
    assert generator.model == model
    assert generator.tokenizer.unwrapped == tokenizer


def test_generate() -> None:
    # arrange
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    mock_tokenizer = MagicMock()
    mock_model = MagicMock()
    mock_model.device = torch.device("cpu")
    mock_model.generate.return_value = torch.Tensor([1, 2, 3])
    mock_tokenizer_result = MagicMock()
    mock_tokenizer_result.input_ids = torch.ones(2)
    mock_tokenizer.batch_decode.return_value = ["Mock output"]
    mock_tokenizer.return_value = mock_tokenizer_result
    generator.tokenizer.unwrapped = mock_tokenizer
    generator.model = mock_model

    # act
    result = generator.generate("fake input", "fake context")

    assert result == "Mock output"
    mock_tokenizer.assert_called_once()
    mock_model.generate.assert_called_once()


@patch("fed_rag.generators.huggingface.mixin.F")
def test_compute_target_sequence_proba(
    mock_torch_functional: MagicMock,
) -> None:
    # arrange
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    mock_tokenizer = MagicMock()
    mock_tokenizer.encode.side_effect = [
        EncodeResult(input_ids=[0, 1, 2, 3, 4]),
        EncodeResult(input_ids=[0, 1, 2]),
    ]
    mock_torch_functional.softmax.return_value = torch.tensor(
        [0.2, 0.2, 0.2, 0.2, 0.2]
    )
    mock_model = MagicMock()
    mock_model.device = torch.device("cpu")
    generator.model = mock_model
    generator.tokenizer = mock_tokenizer

    # act
    result = generator.compute_target_sequence_proba(
        prompt="fake prompt", target=" fake target"
    )

    mock_tokenizer.encode.assert_any_call("fake prompt fake target")
    mock_tokenizer.encode.assert_any_call("fake prompt")
    assert mock_torch_functional.softmax.call_count == 2
    mock_model.assert_called_once()
    assert_close(result, torch.tensor(0.2**2))


def test_huggingface_extra_missing() -> None:
    """Test extra is not installed."""

    modules = {"transformers": None}
    module_to_import = "fed_rag.generators.huggingface.hf_peft_model"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HFPeftModelGenerator` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.generators.huggingface.hf_peft_model import (
                HFPeftModelGenerator,
            )

            HFPeftModelGenerator("fake_name", "fake_base_name")

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module


def test_huggingface_extra_missing_imported_from_parent() -> None:
    """Test extra is not installed."""

    modules = {"transformers": None}
    module_to_import = "fed_rag.generators.huggingface"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`fed_rag.generators.huggingface` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.generators.huggingface import HFPeftModelGenerator

            HFPeftModelGenerator("fake_name", "fake_base_name")

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module



from fed_rag.base.generator import BaseGenerator


def test_generate(mock_generator: BaseGenerator) -> None:
    output = mock_generator.generate("hello")
    assert output == "mock output from 'hello'."


def test_compute_target_sequence_proba(mock_generator: BaseGenerator) -> None:
    proba = mock_generator.compute_target_sequence_proba(
        prompt="mock prompt", target="mock target"
    )
    assert proba == 0.42



import uuid
from abc import ABC, abstractmethod

from pydantic import BaseModel, Field
from typing_extensions import Self


def generate_ks_id() -> str:
    return str(uuid.uuid4())


class ManagedMixin(BaseModel, ABC):
    ks_id: str = Field(default_factory=generate_ks_id)

    @classmethod
    @abstractmethod
    def from_name_and_id(cls, ks_id: str) -> Self:
        """Load a managed Knowledge Store by id."""



"""PyTorch Trainer Config"""

import torch
from torch.utils.data import DataLoader

from fed_rag.base.trainer_config import BaseTrainerConfig


class PyTorchTrainerConfig(BaseTrainerConfig):
    net: torch.nn.Module
    train_data: DataLoader
    val_data: DataLoader



"""Base Bridge"""

import importlib
import importlib.util
from typing import Any, ClassVar, Optional, TypedDict

from pydantic import BaseModel, ConfigDict

from fed_rag.exceptions import (
    MissingExtraError,
    MissingSpecifiedConversionMethod,
)


class BridgeMetadata(TypedDict):
    bridge_version: str
    framework: str
    compatible_versions: list[str]
    method_name: str


class BaseBridgeMixin(BaseModel):
    """Base Bridge Class."""

    # Version of the bridge implementaiton
    _bridge_version: ClassVar[str]
    _bridge_extra: ClassVar[Optional[str | None]]
    _framework: ClassVar[str]
    _compatible_versions: ClassVar[list[str]]
    _method_name: ClassVar[str]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init_subclass__(cls, **kwargs: Any):
        """Register bridge into ~RAGSystem bridge registry."""
        super().__init_subclass__(**kwargs)

        # Register this bridge's metadata to the parent RAGSystem
        for base in cls.__mro__:
            if base.__name__ == "RAGSystem" and hasattr(
                base, "_register_bridge"
            ):
                metadata = cls.get_bridge_metadata()

                # validate method exists
                if not hasattr(cls, metadata["method_name"]):
                    raise MissingSpecifiedConversionMethod(
                        f"Bridge mixin for `{metadata['framework']}` is missing conversion method `{metadata['method_name']}`."
                    )
                base._register_bridge(metadata)
                break

    @classmethod
    def get_bridge_metadata(cls) -> BridgeMetadata:
        metadata: BridgeMetadata = {
            "bridge_version": cls._bridge_version,
            "framework": cls._framework,
            "compatible_versions": cls._compatible_versions,
            "method_name": cls._method_name,
        }
        return metadata

    @classmethod
    def _validate_framework_installed(cls) -> None:
        if importlib.util.find_spec(cls._framework.replace("-", "_")) is None:
            missing_package_or_extra = (
                f"fed-rag[{cls._bridge_extra}]"
                if cls._bridge_extra
                else cls._framework
            )
            msg = (
                f"`{cls._framework}` module is missing but needs to be installed. "
                f"To fix please run `pip install {missing_package_or_extra}`."
            )
            raise MissingExtraError(msg)



"""Data utils"""

from enum import Enum
from typing import Any, Sequence

import torch
from typing_extensions import assert_never

from fed_rag.types.rag_system import RAGSystem
from fed_rag.utils.data.finetuning_datasets import PyTorchRAGFinetuningDataset

DEFAULT_FINETUNE_EXAMPLE_TEMPLATE = "{query} {context} {answer}"


class ReturnType(str, Enum):
    PYTORCH = "pt"
    HUGGINGFACE = "hf"
    TEXT = "txt"


def build_finetune_dataset(
    rag_system: RAGSystem,
    examples: Sequence[dict],
    eos_token_id: int,
    finetune_example_template: str = DEFAULT_FINETUNE_EXAMPLE_TEMPLATE,
    query_key: str = "query",
    answer_key: str = "answer",
    return_dataset: ReturnType = ReturnType.PYTORCH,
) -> Any:
    """Generates the finetuning dataset using the supplied rag_system and examples."""

    if (
        isinstance(return_dataset, str)
        and return_dataset not in ReturnType._value2member_map_.keys()
    ):
        raise ValueError(
            "Invalid `return_type` specified."
        )  # TODO: give a proper exception to this

    inputs_list = []
    targets_list = []
    attention_mask_list = []
    finetuning_instances = []
    for example in examples:
        # retrieve
        source_nodes = rag_system.retrieve(query=example[query_key])
        total_sum_scores = sum(s.score for s in source_nodes)

        # parallel in-context retrieval-augmentation creates
        # top_k separated finetuning instances
        for source in source_nodes:
            finetune_instance_text = finetune_example_template.format(
                query=example[query_key],
                answer=example[answer_key],
                context=source.node.get_content()["text_content"],
            )
            finetuning_instances.append(finetune_instance_text)
            _weight = source.score / total_sum_scores

            # tokenize to get input_ids and target_ids
            tokenizer = rag_system.generator.tokenizer
            encode_result = tokenizer.encode(finetune_instance_text)
            input_ids = encode_result["input_ids"]
            attention_mask = encode_result["attention_mask"]
            target_ids = input_ids[1:] + [eos_token_id]

            inputs_list.append(input_ids)
            targets_list.append(target_ids)
            attention_mask_list.append(attention_mask)

    if return_dataset == ReturnType.TEXT:
        return finetuning_instances
    elif return_dataset == ReturnType.PYTORCH:
        return PyTorchRAGFinetuningDataset(
            input_ids=[torch.Tensor(el) for el in inputs_list],
            target_ids=[torch.Tensor(el) for el in targets_list],
        )
    elif return_dataset == ReturnType.HUGGINGFACE:
        # needs `fed-rag[huggingface]` extra to be installed
        # this import will fail if not installed
        from fed_rag.utils.data.finetuning_datasets.huggingface import (
            HuggingFaceRAGFinetuningDataset,
        )

        return HuggingFaceRAGFinetuningDataset.from_inputs(
            input_ids=inputs_list,
            target_ids=targets_list,
            attention_mask=attention_mask_list,
        )
    else:
        assert_never(return_dataset)  # pragma: no cover



from .core import FedRAGError


class TrainerError(FedRAGError):
    """Base errors for all rag trainer relevant exceptions."""

    pass


class InconsistentDatasetError(TrainerError):
    pass


class InvalidLossError(TrainerError):
    pass


class InvalidDataCollatorError(TrainerError):
    pass


class MissingInputTensor(TrainerError):
    pass



"""PyTorchTrainerConfig Unit Tests"""

from typing import Any

import numpy as np
import pytest
import torch
from torch.utils.data import DataLoader, Dataset

from fed_rag.trainer_configs import PyTorchTrainerConfig


class _TestDataset(Dataset):
    def __init__(self, size: int) -> None:
        self.features = np.random.rand(size, 2)
        self.labels = np.random.choice(2, size=size)

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> tuple[np.ndarray, Any]:
        return self.features[index], self.labels[index]


@pytest.fixture()
def train_dataloader() -> DataLoader:
    dataset = _TestDataset(size=10)
    return DataLoader(dataset, batch_size=2, shuffle=True)


@pytest.fixture()
def val_dataloader() -> DataLoader:
    dataset = _TestDataset(size=4)
    return DataLoader(dataset, batch_size=2, shuffle=True)


def test_init(
    train_dataloader: DataLoader, val_dataloader: DataLoader
) -> None:
    mdl = torch.nn.Linear(2, 1)
    cfg = PyTorchTrainerConfig(
        net=mdl,
        train_data=train_dataloader,
        val_data=val_dataloader,
        a=1,
        b=2,
        c="3",
    )

    assert cfg["a"] == 1
    assert cfg["b"] == 2
    assert cfg["c"] == "3"



from typing import Any

import pytest
import tokenizers
from tokenizers import Tokenizer, models
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from fed_rag.base.tokenizer import BaseTokenizer


class MockTokenizer(BaseTokenizer):
    def encode(self, input: str, **kwargs: Any) -> list[int]:
        return [0, 1, 2]

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return "mock decoded sentence"

    @property
    def unwrapped(self) -> None:
        return None


@pytest.fixture()
def mock_tokenizer() -> BaseTokenizer:
    return MockTokenizer()


@pytest.fixture
def hf_tokenizer() -> PreTrainedTokenizer:
    tokenizer = Tokenizer(
        models.WordPiece({"hello": 0, "[UNK]": 1}, unk_token="[UNK]")
    )
    tokenizer.pre_tokenizer = tokenizers.pre_tokenizers.WhitespaceSplit()
    return PreTrainedTokenizerFast(
        tokenizer_object=tokenizer,
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]",
    )



"""HuggingFace Retrieval-Augmented Generator Trainer"""

from typing import TYPE_CHECKING, Any, Optional

from pydantic import PrivateAttr, model_validator

from fed_rag.base.trainer import BaseGeneratorTrainer
from fed_rag.data_collators.huggingface.ralt import DataCollatorForRALT
from fed_rag.exceptions import MissingExtraError
from fed_rag.trainers.huggingface.mixin import HuggingFaceTrainerMixin
from fed_rag.types.rag_system import RAGSystem
from fed_rag.types.results import TestResult, TrainResult
from fed_rag.utils.huggingface import _validate_rag_system

try:
    from transformers import Trainer

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

    class Trainer:  # type: ignore[no-redef]
        """Dummy placeholder when transformers is not available."""

        pass


if TYPE_CHECKING:  # pragma: no cover
    from datasets import Dataset
    from transformers import Trainer, TrainingArguments
    from transformers.trainer_utils import TrainOutput


def _get_default_training_args() -> "TrainingArguments":
    from transformers import TrainingArguments

    return TrainingArguments(remove_unused_columns=False)


class HuggingFaceTrainerForRALT(HuggingFaceTrainerMixin, BaseGeneratorTrainer):
    """HuggingFace Trainer for Retrieval-Augmented LM Training/Fine-Tuning."""

    _hf_trainer: Optional["Trainer"] = PrivateAttr(default=None)

    def __init__(
        self,
        rag_system: RAGSystem,
        train_dataset: "Dataset",
        training_arguments: Optional["TrainingArguments"] = None,
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        if training_arguments is None:
            training_arguments = _get_default_training_args()
        else:
            training_arguments.remove_unused_columns = (
                False  # pragma: no cover
            )

        super().__init__(
            train_dataset=train_dataset,
            rag_system=rag_system,
            training_arguments=training_arguments,
            **kwargs,
        )

    @model_validator(mode="after")
    def set_private_attributes(self) -> "HuggingFaceTrainerForRALT":
        # if made it to here, then this import is available
        from transformers import Trainer

        # validate rag system
        _validate_rag_system(self.rag_system)

        self._hf_trainer = Trainer(
            model=self.model,
            args=self.training_arguments,
            data_collator=DataCollatorForRALT(rag_system=self.rag_system),
            train_dataset=self.train_dataset,
        )

        return self

    def train(self) -> TrainResult:
        output: TrainOutput = self.hf_trainer_obj.train()
        return TrainResult(loss=output.training_loss)

    def evaluate(self) -> TestResult:
        # TODO: implement this
        raise NotImplementedError

    @property
    def hf_trainer_obj(self) -> "Trainer":
        return self._hf_trainer



"""Exceptions for Bridges."""

from .core import FedRAGError


class BridgeError(FedRAGError):
    """Base bridge error for all fl-task-related exceptions."""

    pass


class MissingSpecifiedConversionMethod(BridgeError):
    pass



"""PyTorch Trainer Mixin"""

from abc import ABC
from typing import Any, Protocol, runtime_checkable

import torch.nn as nn
from pydantic import BaseModel, ConfigDict
from torch.utils.data import DataLoader, Dataset

from fed_rag.exceptions import InconsistentDatasetError

from .training_args import TrainingArgs


# Define the protocol for runtime checking
@runtime_checkable
class PyTorchTrainerProtocol(Protocol):
    train_dataset: Dataset
    training_arguments: TrainingArgs | None
    train_dataloader: DataLoader

    def model(self) -> nn.Module:
        pass  # pragma: no cover


class PyTorchTrainerMixin(BaseModel, ABC):
    """HuggingFace Trainer Mixin."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )
    train_dataset: Dataset
    train_dataloader: DataLoader
    training_arguments: TrainingArgs | None = None

    def __init__(
        self,
        train_dataloader: DataLoader,
        train_dataset: Dataset | None = None,
        training_arguments: TrainingArgs | None = None,
        **kwargs: Any,
    ):
        if train_dataset is None:
            train_dataset = train_dataloader.dataset
        else:
            # ensure consistency between loader.dataset and the supplied one
            if id(train_dataset) != id(train_dataloader.dataset):
                raise InconsistentDatasetError(
                    "Inconsistent datasets detected between supplied `train_dataset` and that "
                    "associated with the `train_dataloader`. These two datasets must be the same."
                )

        super().__init__(
            train_dataset=train_dataset,
            train_dataloader=train_dataloader,
            training_arguments=training_arguments,
            **kwargs,
        )



"""HuggingFace LM-Supervised Retriever Trainer"""

from typing import TYPE_CHECKING, Any, Optional

import torch
from pydantic import PrivateAttr, model_validator

from fed_rag.base.trainer import BaseRetrieverTrainer
from fed_rag.data_collators.huggingface import DataCollatorForLSR
from fed_rag.exceptions import (
    InvalidDataCollatorError,
    InvalidLossError,
    MissingExtraError,
    MissingInputTensor,
    TrainerError,
)
from fed_rag.loss.pytorch.lsr import LSRLoss
from fed_rag.trainers.huggingface.mixin import HuggingFaceTrainerMixin
from fed_rag.types.rag_system import RAGSystem
from fed_rag.types.results import TestResult, TrainResult
from fed_rag.utils.huggingface import _validate_rag_system

try:
    from sentence_transformers import SentenceTransformerTrainer

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

    class SentenceTransformerTrainer:  # type: ignore[no-redef]
        """Dummy placeholder when sentence transformers is not available."""

        pass


if TYPE_CHECKING:  # pragma: no cover
    from datasets import Dataset
    from sentence_transformers import (
        SentenceTransformer,
        SentenceTransformerTrainer,
    )
    from transformers import TrainingArguments
    from transformers.trainer_utils import TrainOutput


class LSRSentenceTransformerTrainer(SentenceTransformerTrainer):
    def __init__(
        self,
        *args: Any,
        data_collator: DataCollatorForLSR,
        loss: Optional[LSRLoss] = None,
        **kwargs: Any,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        # set loss
        if loss is None:
            loss = LSRLoss()
        else:
            if not isinstance(loss, LSRLoss):
                raise InvalidLossError(
                    "`LSRSentenceTransformerTrainer` must use ~fed_rag.loss.LSRLoss`."
                )

        if not isinstance(data_collator, DataCollatorForLSR):
            raise InvalidDataCollatorError(
                "`LSRSentenceTransformerTrainer` must use ~fed_rag.data_collators.DataCollatorForLSR`."
            )

        super().__init__(
            *args, loss=loss, data_collator=data_collator, **kwargs
        )

    def collect_scores(
        self, inputs: dict[str, torch.Tensor | Any]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if "retrieval_scores" not in inputs:
            raise MissingInputTensor(
                "Collated `inputs` are missing key `retrieval_scores`"
            )

        if "lm_scores" not in inputs:
            raise MissingInputTensor(
                "Collated `inputs` are missing key `lm_scores`"
            )

        retrieval_scores = inputs.get("retrieval_scores")
        lm_scores = inputs.get("lm_scores")

        return retrieval_scores, lm_scores

    def compute_loss(
        self,
        model: "SentenceTransformer",
        inputs: dict[str, torch.Tensor | Any],
        return_outputs: bool = False,
        num_items_in_batch: Any | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, dict[str, Any]]:
        """Compute LSR loss.

        NOTE: the forward pass of the model is taken care of in the DataCollatorForLSR.

        Args:
            model (SentenceTransformer): _description_
            inputs (dict[str, torch.Tensor  |  Any]): _description_
            return_outputs (bool, optional): _description_. Defaults to False.
            num_items_in_batch (Any | None, optional): _description_. Defaults to None.

        Raises:
            NotImplementedError: _description_

        Returns:
            torch.Tensor | tuple[torch.Tensor, dict[str, Any]]: _description_
        """
        retrieval_scores, lm_scores = self.collect_scores(inputs)
        loss = self.loss(retrieval_scores, lm_scores)

        # inputs are actually the outputs of RAGSystem's "forward" pass
        return (loss, inputs) if return_outputs else loss


class HuggingFaceTrainerForLSR(HuggingFaceTrainerMixin, BaseRetrieverTrainer):
    """HuggingFace LM-Supervised Retriever Trainer."""

    _hf_trainer: Optional["SentenceTransformerTrainer"] = PrivateAttr(
        default=None
    )

    def __init__(
        self,
        rag_system: RAGSystem,
        train_dataset: "Dataset",
        training_arguments: Optional["TrainingArguments"] = None,
        **kwargs: Any,
    ):
        super().__init__(
            train_dataset=train_dataset,
            rag_system=rag_system,
            training_arguments=training_arguments,
            **kwargs,
        )

    @model_validator(mode="after")
    def set_private_attributes(self) -> "HuggingFaceTrainerForLSR":
        # if made it to here, then this import is available
        from sentence_transformers import SentenceTransformer

        # validate rag system
        _validate_rag_system(self.rag_system)

        # validate model
        if not isinstance(self.model, SentenceTransformer):
            raise TrainerError(
                "For `HuggingFaceTrainerForLSR`, attribute `model` must be of type "
                "`~sentence_transformers.SentenceTransformer`."
            )

        self._hf_trainer = LSRSentenceTransformerTrainer(
            model=self.model,
            args=self.training_arguments,
            data_collator=DataCollatorForLSR(rag_system=self.rag_system),
            train_dataset=self.train_dataset,
        )

        return self

    def train(self) -> TrainResult:
        output: TrainOutput = self.hf_trainer_obj.train()
        return TrainResult(loss=output.training_loss)

    def evaluate(self) -> TestResult:
        # TODO: implement this
        raise NotImplementedError

    @property
    def hf_trainer_obj(self) -> "SentenceTransformerTrainer":
        return self._hf_trainer



import re
import sys
from unittest.mock import patch

import pytest
from datasets import Dataset
from pytest import MonkeyPatch
from transformers import Trainer, TrainingArguments

from fed_rag.base.trainer import BaseRetrieverTrainer, BaseTrainer
from fed_rag.exceptions import MissingExtraError
from fed_rag.trainers.huggingface.mixin import HuggingFaceTrainerProtocol
from fed_rag.types.rag_system import RAGSystem

from .conftest import TestHFGeneratorTrainer


def test_hf_trainer_init(
    train_dataset: Dataset, hf_rag_system: RAGSystem, monkeypatch: MonkeyPatch
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    # arrange
    training_args = TrainingArguments()
    trainer = TestHFGeneratorTrainer(
        rag_system=hf_rag_system,
        train_dataset=train_dataset,
        training_arguments=training_args,
    )

    assert trainer.rag_system == hf_rag_system
    assert trainer.model == hf_rag_system.generator.model
    assert trainer.train_dataset == train_dataset
    assert trainer.training_arguments == training_args
    assert isinstance(trainer, HuggingFaceTrainerProtocol)
    assert isinstance(trainer, BaseTrainer)


def test_huggingface_extra_missing(
    train_dataset: Dataset, hf_rag_system: RAGSystem, monkeypatch: MonkeyPatch
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    modules = {
        "transformers": None,
    }
    module_to_import = "fed_rag.trainers.huggingface.mixin"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`TestHFRetrieverTrainer` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.trainers.huggingface.mixin import (
                HuggingFaceTrainerMixin,
            )
            from fed_rag.types.results import TestResult, TrainResult

            class TestHFRetrieverTrainer(
                HuggingFaceTrainerMixin, BaseRetrieverTrainer
            ):
                __test__ = False  # needed for Pytest collision. Avoids PytestCollectionWarning

                def train(self) -> TrainResult:
                    return TrainResult(loss=0.42)

                def evaluate(self) -> TestResult:
                    return TestResult(loss=0.42)

                def hf_trainer_obj(self) -> Trainer:
                    return Trainer()

            training_args = TrainingArguments()
            TestHFRetrieverTrainer(
                rag_system=hf_rag_system,
                train_dataset=train_dataset,
                training_arguments=training_args,
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module



"""PyTorchFLTask Unit Tests"""

from typing import Any, Callable

import pytest
import torch

# huggingface
from datasets import Dataset
from peft import LoraConfig, PeftModel, get_peft_model
from sentence_transformers import SentenceTransformer
from transformers import PretrainedConfig, PreTrainedModel

from fed_rag.decorators import federate
from fed_rag.types import TestResult, TrainResult


@pytest.fixture
def hf_pretrained_model() -> PreTrainedModel:
    class _TestHFConfig(PretrainedConfig):
        model_type = "testmodel"

        def __init__(self, num_hidden: int = 42, **kwargs: Any):
            super().__init__(**kwargs)
            self.num_hidden = num_hidden

    class _TestHFPretrainedModel(PreTrainedModel):
        config_class = _TestHFConfig

        def __init__(self, config: _TestHFConfig):
            super().__init__(config)
            self.config = config
            self.model = torch.nn.Linear(3, self.config.num_hidden)

        def forward(self, input: torch.Tensor) -> torch.Tensor:
            return self.model(input)

    return _TestHFPretrainedModel(_TestHFConfig())


@pytest.fixture
def hf_sentence_transformer() -> SentenceTransformer:
    return SentenceTransformer(modules=[torch.nn.Linear(5, 5)])


@pytest.fixture
def hf_peft_model() -> PeftModel:
    class MLP(torch.nn.Module):
        """Taken from Peft's documentation, specifially 'Custom Models'.

        https://huggingface.co/docs/peft/main/en/developer_guides/custom_models
        """

        def __init__(self, num_units_hidden: int = 40):
            super().__init__()
            self.seq = torch.nn.Sequential(
                torch.nn.Linear(20, num_units_hidden),
                torch.nn.ReLU(),
                torch.nn.Linear(num_units_hidden, num_units_hidden),
                torch.nn.ReLU(),
                torch.nn.Linear(num_units_hidden, 2),
                torch.nn.LogSoftmax(dim=-1),
            )

        def forward(self, X: torch.Tensor) -> torch.Tensor:
            return self.seq(X)

    config = LoraConfig(
        target_modules=["seq.0", "seq.2"],
        r=4,
    )

    base_model = MLP()
    return get_peft_model(base_model, config)


# Datasets
@pytest.fixture
def train_dataset() -> Dataset:
    return Dataset.from_dict(
        {
            "text1": ["My first sentence", "Another pair"],
            "text2": ["My second sentence", "Unrelated sentence"],
            "label": [0.8, 0.3],
        }
    )


@pytest.fixture
def val_dataset() -> Dataset:
    return Dataset.from_dict(
        {
            "text1": ["My third sentence"],
            "text2": ["My fourth sentence"],
            "label": [
                0.42,
            ],
        }
    )


# trainers and testers
@pytest.fixture()
def trainer_pretrained_model() -> Callable:
    @federate.trainer.huggingface
    def fn(
        net: PreTrainedModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
    ) -> TrainResult:
        return TrainResult(loss=0.0)

    return fn  # type: ignore


@pytest.fixture()
def tester_pretrained_model() -> Callable:
    @federate.tester.huggingface
    def fn(
        net: PreTrainedModel,
        test_dataset: Dataset,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore


@pytest.fixture()
def mismatch_tester_pretrained_model() -> Callable:
    @federate.tester.huggingface
    def fn(
        mdl: PreTrainedModel,  # mismatch here
        test_dataset: Dataset,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore


@pytest.fixture()
def undecorated_trainer() -> Callable:
    def fn(
        net: PreTrainedModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
    ) -> TrainResult:
        return TrainResult(loss=0.0)

    return fn  # type: ignore


@pytest.fixture()
def undecorated_tester() -> Callable:
    def fn(
        net: PreTrainedModel,
        test_dataset: Dataset,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore


@pytest.fixture()
def trainer_sentence_transformer() -> Callable:
    @federate.trainer.huggingface
    def fn(
        net: SentenceTransformer,
        train_dataset: Dataset,
        val_dataset: Dataset,
    ) -> TrainResult:
        return TrainResult(loss=0.0)

    return fn  # type: ignore


@pytest.fixture()
def tester_sentence_transformer() -> Callable:
    @federate.tester.huggingface
    def fn(
        net: SentenceTransformer,
        test_dataset: Dataset,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore


@pytest.fixture()
def trainer_peft_model() -> Callable:
    @federate.trainer.huggingface
    def fn(
        net: PeftModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
    ) -> TrainResult:
        return TrainResult(loss=0.0)

    return fn  # type: ignore


@pytest.fixture()
def tester_peft_model() -> Callable:
    @federate.tester.huggingface
    def fn(
        net: PeftModel,
        test_dataset: Dataset,
    ) -> TestResult:
        return TestResult(loss=0.0, metrics={})

    return fn  # type: ignore



"""Decorators unit tests for HuggingFace"""

from typing import Any

import pytest
from datasets import Dataset
from peft import PeftModel
from sentence_transformers import SentenceTransformer
from transformers import PreTrainedModel

from fed_rag.decorators import federate
from fed_rag.exceptions.inspectors import (
    InvalidReturnType,
    MissingDataParam,
    MissingMultipleDataParams,
    MissingNetParam,
)
from fed_rag.inspectors.huggingface import (
    TesterSignatureSpec,
    TrainerSignatureSpec,
)
from fed_rag.types import TestResult, TrainResult


## For All
def test_decorated_trainer_raises_missing_net_param_error() -> None:
    def train_loop(
        train_dataset: Dataset,
        val_dataset: Dataset,
    ) -> TrainResult:
        pass

    with pytest.raises(MissingNetParam):
        federate.trainer.huggingface(train_loop)


## PretrainedModel
### Trainer
def test_decorated_trainer_hf_pretrained_model() -> None:
    def train_loop(
        net: PreTrainedModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TrainResult:
        pass

    decorated = federate.trainer.huggingface(train_loop)
    config: TrainerSignatureSpec = getattr(
        decorated, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_dataset"
    assert config.val_data_param == "val_dataset"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PreTrainedModel"


def test_decorated_trainer_raises_invalid_return_type_error_hf_pretrained() -> (
    None
):
    def train_loop(
        net: PreTrainedModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Trainer should return a fed_rag.types.TrainResult or a subclass of it.",
    ):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_to_find_two_data_params_hf_pretrained() -> (
    None
):
    def train_loop(
        model: PreTrainedModel,
    ) -> TrainResult:
        pass

    msg = (
        "Inspection failed to find two data params for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingMultipleDataParams, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_due_to_missing_data_loader_hf_pretrained() -> (
    None
):
    def train_loop(model: PreTrainedModel, val_loader: Dataset) -> TrainResult:
        pass

    msg = (
        "Inspection found one data param but failed to find another. "
        "Two data params are required for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_from_instance_method_hf_pretrained() -> None:
    class _TestClass:
        @federate.trainer.huggingface
        def train_loop(
            self,
            net: PreTrainedModel,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PreTrainedModel"


def test_decorated_trainer_from_class_method_hf_pretrained() -> None:
    class _TestClass:
        @classmethod
        @federate.trainer.huggingface
        def train_loop(
            cls,
            net: PreTrainedModel,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PreTrainedModel"


### Tester
def test_decorated_tester_hf_pretrained() -> None:
    def fn(
        mdl: PreTrainedModel,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TestResult:
        pass

    decorated = federate.tester.huggingface(fn)
    config: TesterSignatureSpec = getattr(decorated, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PreTrainedModel"


def test_decorated_tester_raises_invalid_return_type_hf_pretrained() -> None:
    def fn(
        mdl: PreTrainedModel,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Tester should return a fed_rag.types.TestResult or a subclass of it.",
    ):
        federate.tester.huggingface(fn)


def test_decorated_tester_raises_missing_net_param_error_hf_pretrained() -> (
    None
):
    def fn(
        test_loader: Dataset,
    ) -> TestResult:
        pass

    with pytest.raises(MissingNetParam):
        federate.tester.huggingface(fn)


def test_decorated_tester_fails_to_find_a_data_params_hf_pretrained() -> None:
    def fn(
        model: PreTrainedModel,
    ) -> TestResult:
        pass

    msg = (
        "Inspection failed to find a data param for a test dataset."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.tester.huggingface(fn)


def test_decorated_tester_from_instance_method_hf_pretrained() -> None:
    class _TestClass:
        @federate.tester.huggingface
        def fn(
            self,
            mdl: PreTrainedModel,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    obj = _TestClass()
    config: TesterSignatureSpec = getattr(obj.fn, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "PreTrainedModel"


def test_decorated_tester_from_class_method_hf_pretrained() -> None:
    class _TestClass:
        @classmethod
        @federate.tester.huggingface
        def fn(
            cls,
            mdl: PreTrainedModel,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    config: TesterSignatureSpec = getattr(
        _TestClass.fn, "__fl_task_tester_config"
    )
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "PreTrainedModel"


## Sentence Transformers
### Trainer
def test_decorated_trainer_hf_st() -> None:
    def train_loop(
        net: SentenceTransformer,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TrainResult:
        pass

    decorated = federate.trainer.huggingface(train_loop)
    config: TrainerSignatureSpec = getattr(
        decorated, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_dataset"
    assert config.val_data_param == "val_dataset"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "SentenceTransformer"


def test_decorated_trainer_raises_invalid_return_type_error_hf_st() -> None:
    def train_loop(
        net: SentenceTransformer,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Trainer should return a fed_rag.types.TrainResult or a subclass of it.",
    ):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_to_find_two_data_params_hf_st() -> None:
    def train_loop(
        model: SentenceTransformer,
    ) -> TrainResult:
        pass

    msg = (
        "Inspection failed to find two data params for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingMultipleDataParams, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_due_to_missing_data_loader_hf_st() -> None:
    def train_loop(
        model: SentenceTransformer, val_loader: Dataset
    ) -> TrainResult:
        pass

    msg = (
        "Inspection found one data param but failed to find another. "
        "Two data params are required for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_from_instance_method_hf_st() -> None:
    class _TestClass:
        @federate.trainer.huggingface
        def train_loop(
            self,
            net: SentenceTransformer,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "SentenceTransformer"


def test_decorated_trainer_from_class_method_hf_st() -> None:
    class _TestClass:
        @classmethod
        @federate.trainer.huggingface
        def train_loop(
            cls,
            net: SentenceTransformer,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "SentenceTransformer"


### Tester
def test_decorated_tester_hf_st() -> None:
    def fn(
        mdl: SentenceTransformer,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TestResult:
        pass

    decorated = federate.tester.huggingface(fn)
    config: TesterSignatureSpec = getattr(decorated, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "SentenceTransformer"


def test_decorated_tester_raises_invalid_return_type_hf_st() -> None:
    def fn(
        mdl: SentenceTransformer,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Tester should return a fed_rag.types.TestResult or a subclass of it.",
    ):
        federate.tester.huggingface(fn)


def test_decorated_tester_fails_to_find_a_data_params_hf_st() -> None:
    def fn(
        model: SentenceTransformer,
    ) -> TestResult:
        pass

    msg = (
        "Inspection failed to find a data param for a test dataset."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.tester.huggingface(fn)


def test_decorated_tester_from_instance_method_hf_st() -> None:
    class _TestClass:
        @federate.tester.huggingface
        def fn(
            self,
            mdl: SentenceTransformer,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    obj = _TestClass()
    config: TesterSignatureSpec = getattr(obj.fn, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "SentenceTransformer"


def test_decorated_tester_from_class_method_hf_st() -> None:
    class _TestClass:
        @classmethod
        @federate.tester.huggingface
        def fn(
            cls,
            mdl: SentenceTransformer,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    config: TesterSignatureSpec = getattr(
        _TestClass.fn, "__fl_task_tester_config"
    )
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "SentenceTransformer"


## PeftModel
### Trainer
def test_decorated_trainer_hf_peft() -> None:
    def train_loop(
        net: PeftModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TrainResult:
        pass

    decorated = federate.trainer.huggingface(train_loop)
    config: TrainerSignatureSpec = getattr(
        decorated, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_dataset"
    assert config.val_data_param == "val_dataset"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PeftModel"


def test_decorated_trainer_raises_invalid_return_type_error_hf_peft() -> None:
    def train_loop(
        net: PeftModel,
        train_dataset: Dataset,
        val_dataset: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Trainer should return a fed_rag.types.TrainResult or a subclass of it.",
    ):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_to_find_two_data_params_hf_peft() -> None:
    def train_loop(
        model: PeftModel,
    ) -> TrainResult:
        pass

    msg = (
        "Inspection failed to find two data params for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingMultipleDataParams, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_fails_due_to_missing_data_loader_hf_peft() -> None:
    def train_loop(model: PeftModel, val_loader: Dataset) -> TrainResult:
        pass

    msg = (
        "Inspection found one data param but failed to find another. "
        "Two data params are required for train and val datasets."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.trainer.huggingface(train_loop)


def test_decorated_trainer_from_instance_method_hf_peft() -> None:
    class _TestClass:
        @federate.trainer.huggingface
        def train_loop(
            self,
            net: PeftModel,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PeftModel"


def test_decorated_trainer_from_class_method_hf_peft() -> None:
    class _TestClass:
        @classmethod
        @federate.trainer.huggingface
        def train_loop(
            cls,
            net: PeftModel,
            train_loader: Dataset,
            val_loader: Dataset,
            extra_param_1: int,
            extra_param_2: float | None,
        ) -> TrainResult:
            pass

    obj = _TestClass()
    config: TrainerSignatureSpec = getattr(
        obj.train_loop, "__fl_task_trainer_config"
    )
    assert config.net_parameter == "net"
    assert config.train_data_param == "train_loader"
    assert config.val_data_param == "val_loader"
    assert config.extra_train_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PeftModel"


### Tester
def test_decorated_tester_hf_peft() -> None:
    def fn(
        mdl: PeftModel,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> TestResult:
        pass

    decorated = federate.tester.huggingface(fn)
    config: TesterSignatureSpec = getattr(decorated, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1", "extra_param_2"]
    assert config.net_parameter_class_name == "PeftModel"


def test_decorated_tester_raises_invalid_return_type_hf_peft() -> None:
    def fn(
        mdl: PeftModel,
        test_loader: Dataset,
        extra_param_1: int,
        extra_param_2: float | None,
    ) -> Any:
        pass

    with pytest.raises(
        InvalidReturnType,
        match="Tester should return a fed_rag.types.TestResult or a subclass of it.",
    ):
        federate.tester.huggingface(fn)


def test_decorated_tester_fails_to_find_a_data_params_hf_peft() -> None:
    def fn(
        model: PeftModel,
    ) -> TestResult:
        pass

    msg = (
        "Inspection failed to find a data param for a test dataset."
        "For HuggingFace these params must be of type `datasets.Dataset`"
    )
    with pytest.raises(MissingDataParam, match=msg):
        federate.tester.huggingface(fn)


def test_decorated_tester_from_instance_method_hf_peft() -> None:
    class _TestClass:
        @federate.tester.huggingface
        def fn(
            self,
            mdl: PeftModel,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    obj = _TestClass()
    config: TesterSignatureSpec = getattr(obj.fn, "__fl_task_tester_config")
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "PeftModel"


def test_decorated_tester_from_class_method_hf_peft() -> None:
    class _TestClass:
        @classmethod
        @federate.tester.huggingface
        def fn(
            cls,
            mdl: PeftModel,
            test_loader: Dataset,
            extra_param_1: int,
        ) -> TestResult:
            pass

    config: TesterSignatureSpec = getattr(
        _TestClass.fn, "__fl_task_tester_config"
    )
    assert config.net_parameter == "mdl"
    assert config.test_data_param == "test_loader"
    assert config.extra_test_kwargs == ["extra_param_1"]
    assert config.net_parameter_class_name == "PeftModel"



"""Base Trainer Config"""

from typing import Any, Dict, cast

from pydantic import BaseModel, ConfigDict, PrivateAttr, model_serializer


class BaseTrainerConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    net: Any
    train_data: Any
    val_data: Any
    _extra_train_kwargs: Dict[str, Any] = PrivateAttr(
        default_factory=dict
    )  # additional kwargs

    def __init__(self, **params: Any):
        """__init__.

        Sets specified fields and private attrs of the TrainerConfig and then
        stores any additional passed params in _extra_train_kwargs.
        """
        fields = {}
        private_attrs = {}
        extra_train_kwargs = {}
        for k, v in params.items():
            if k in self.model_fields:
                fields[k] = v
            elif k in self.__private_attributes__:
                private_attrs[k] = v
            else:
                extra_train_kwargs[k] = v
        super().__init__(**fields)
        for private_attr, value in private_attrs.items():
            super().__setattr__(private_attr, value)
        if extra_train_kwargs:
            self._extra_train_kwargs.update(extra_train_kwargs)

    @model_serializer(mode="wrap")
    def custom_model_dump(self, handler: Any) -> Dict[str, Any]:
        data = handler(self)
        data = cast(Dict[str, Any], data)
        # include _extra_train_kwargs in serialization
        if self._extra_train_kwargs:
            data["_extra_train_kwargs"] = self._extra_train_kwargs
        return data  # type: ignore[no-any-return]

    def __getattr__(self, __name: str) -> Any:
        if (
            __name in self.__private_attributes__
            or __name in self.model_fields
        ):
            return super().__getattr__(__name)  # type: ignore
        else:
            try:
                return self._data[__name]
            except KeyError:
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{__name}'"
                )

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__private_attributes__ or name in self.model_fields:
            super().__setattr__(name, value)
        else:
            self._extra_train_kwargs.__setitem__(name, value)

    def __getitem__(self, key: str) -> Any:
        return self._extra_train_kwargs[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._extra_train_kwargs[key] = value

    def get(self, key: str, default: Any = None) -> Any:
        return self._extra_train_kwargs.get(key, default)

    def __contains__(self, key: str) -> bool:
        return key in self._extra_train_kwargs



from typing import Any

import pytest
import torch
from pydantic import PrivateAttr
from sentence_transformers import SentenceTransformer

from fed_rag.base.generator import BaseGenerator
from fed_rag.base.retriever import BaseRetriever
from fed_rag.base.tokenizer import BaseTokenizer
from fed_rag.types.knowledge_node import KnowledgeNode


class MockRetriever(BaseRetriever):
    _encoder: torch.nn.Module = PrivateAttr(default=torch.nn.Linear(3, 3))

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(3))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(3))

    @property
    def encoder(self) -> torch.nn.Module:
        return self._encoder

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return None


class MockDualRetriever(BaseRetriever):
    _query_encoder: torch.nn.Module = PrivateAttr(
        default=torch.nn.Linear(2, 1)
    )
    _context_encoder: torch.nn.Module = PrivateAttr(
        default=torch.nn.Linear(2, 1)
    )

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(2))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(2))

    @property
    def encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return self._query_encoder

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return self._context_encoder


@pytest.fixture
def mock_retriever() -> MockRetriever:
    return MockRetriever()


@pytest.fixture
def mock_dual_retriever() -> MockDualRetriever:
    return MockDualRetriever()


@pytest.fixture
def dummy_sentence_transformer() -> SentenceTransformer:
    return SentenceTransformer(modules=[torch.nn.Linear(5, 5)])


@pytest.fixture
def knowledge_nodes() -> list[KnowledgeNode]:
    return [
        KnowledgeNode(
            embedding=[1.0, 0.0, 1.0], node_type="text", text_content="node 1"
        ),
        KnowledgeNode(
            embedding=[1.0, 0.0, 0.0],
            node_type="multimodal",
            text_content="node 2",
            image_content=b"node 2",
        ),
        KnowledgeNode(
            embedding=[
                1.0,
                1.0,
                0.0,
            ],
            node_type="multimodal",
            text_content="node 3",
            image_content=b"node 3",
        ),
    ]


class MockTokenizer(BaseTokenizer):
    def encode(self, input: str, **kwargs: Any) -> list[int]:
        return [0, 1, 2]

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return "mock decoded sentence"

    @property
    def unwrapped(self) -> None:
        return None


@pytest.fixture()
def mock_tokenizer() -> BaseTokenizer:
    return MockTokenizer()


class MockGenerator(BaseGenerator):
    _model = torch.nn.Linear(2, 1)
    _tokenizer = MockTokenizer()

    def generate(self, input: str) -> str:
        return f"mock output from '{input}'."

    def compute_target_sequence_proba(self, prompt: str, target: str) -> float:
        return 0.42

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @property
    def tokenizer(self) -> MockTokenizer:
        return self._tokenizer


@pytest.fixture
def mock_generator() -> BaseGenerator:
    return MockGenerator()



"""HuggingFace PeftModel Generator"""

from typing import TYPE_CHECKING, Any, Optional

from pydantic import ConfigDict, Field, PrivateAttr

try:
    from peft import PeftModel, prepare_model_for_kbit_training
    from transformers import AutoModelForCausalLM
    from transformers.generation.utils import GenerationConfig

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False


if TYPE_CHECKING:  # pragma: no cover
    from peft import PeftModel
    from transformers.generation.utils import GenerationConfig

from fed_rag.base.generator import BaseGenerator
from fed_rag.exceptions.common import MissingExtraError
from fed_rag.tokenizers.hf_pretrained_tokenizer import HFPretrainedTokenizer

from .mixin import HuggingFaceGeneratorMixin

DEFAULT_PROMPT_TEMPLATE = """
You are a helpful assistant. Given the user's question, provide a succinct
and accurate response. If context is provided, use it in your answer if it helps
you to create the most accurate response.

<question>
{question}
</question>

<context>
{context}
</context>

<response>

"""


class HFPeftModelGenerator(HuggingFaceGeneratorMixin, BaseGenerator):
    """HFPeftModelGenerator Class.

    NOTE: this class supports loading PeftModel's from HF Hub or from local.
    TODO: support loading custom models via a `~peft.Config` and `~peft.get_peft_model`
    """

    model_config = ConfigDict(protected_namespaces=("pydantic_model_",))
    model_name: str = Field(
        description="Name of Peft model. Used for loading model from HF hub or local."
    )
    base_model_name: str = Field(
        description="Name of the frozen HuggingFace base model. Used for loading the model from HF hub or local."
    )
    generation_config: "GenerationConfig" = Field(
        description="The generation config used for generating with the PreTrainedModel."
    )
    load_model_kwargs: dict = Field(
        description="Optional kwargs dict for loading peft model from HF. Defaults to None.",
        default_factory=dict,
    )
    load_base_model_kwargs: dict = Field(
        description="Optional kwargs dict for loading base model from HF. Defaults to None.",
        default_factory=dict,
    )
    prompt_template: str = Field(description="Prompt template for RAG.")
    _model: Optional["PeftModel"] = PrivateAttr(default=None)
    _tokenizer: HFPretrainedTokenizer | None = PrivateAttr(default=None)

    def __init__(
        self,
        model_name: str,
        base_model_name: str,
        generation_config: Optional["GenerationConfig"] = None,
        prompt_template: str | None = None,
        load_model_kwargs: dict | None = None,
        load_base_model_kwargs: dict | None = None,
        load_model_at_init: bool = True,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        generation_config = (
            generation_config if generation_config else GenerationConfig()
        )
        prompt_template = (
            prompt_template if prompt_template else DEFAULT_PROMPT_TEMPLATE
        )
        super().__init__(
            model_name=model_name,
            base_model_name=base_model_name,
            generation_config=generation_config,
            prompt_template=prompt_template,
            load_model_kwargs=load_model_kwargs if load_model_kwargs else {},
            load_base_model_kwargs=(
                load_base_model_kwargs if load_base_model_kwargs else {}
            ),
        )
        self._tokenizer = HFPretrainedTokenizer(
            model_name=base_model_name, load_model_at_init=load_model_at_init
        )
        if load_model_at_init:
            self._model = self._load_model_from_hf()

    def _load_model_from_hf(self, **kwargs: Any) -> "PeftModel":
        load_base_kwargs = self.load_base_model_kwargs
        load_kwargs = self.load_model_kwargs
        load_kwargs.update(kwargs)
        self.load_model_kwargs = load_kwargs  # update load_model_kwargs
        base_model = AutoModelForCausalLM.from_pretrained(
            self.base_model_name, **load_base_kwargs
        )

        if "quantization_config" in load_base_kwargs:
            # preprocess model for kbit fine-tuning
            # https://huggingface.co/docs/peft/developer_guides/quantization
            base_model = prepare_model_for_kbit_training(base_model)

        return PeftModel.from_pretrained(
            base_model, self.model_name, **load_kwargs
        )

    @property
    def model(self) -> "PeftModel":
        if self._model is None:
            # load HF PeftModel
            self._model = self._load_model_from_hf()
        return self._model

    @model.setter
    def model(self, value: "PeftModel") -> None:
        self._model = value

    @property
    def tokenizer(self) -> HFPretrainedTokenizer:
        return self._tokenizer

    @tokenizer.setter
    def tokenizer(self, value: HFPretrainedTokenizer) -> None:
        self._tokenizer = value



"""Training Args"""

from typing import Any

from pydantic import BaseModel, Field


class TrainingArgs(BaseModel):
    """Arguments for training."""

    learning_rate: float | None = None
    batch_size: int | None = None
    num_epochs: int | None = None
    warmup_steps: int | None = None
    weight_decay: float | None = None
    custom_kwargs: dict[str, Any] = Field(default_factory=dict)



"""HuggingFace SentenceTransformer Retriever"""

from typing import TYPE_CHECKING, Any, Literal, Optional, cast

import torch
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

try:
    from sentence_transformers import SentenceTransformer

    _has_huggingface = True
except ModuleNotFoundError:
    _has_huggingface = False

if TYPE_CHECKING:  # pragma: no cover
    from sentence_transformers import SentenceTransformer

from fed_rag.base.retriever import BaseRetriever
from fed_rag.exceptions import MissingExtraError


class LoadKwargs(BaseModel):
    encoder: dict = Field(default_factory=dict)
    query_encoder: dict = Field(default_factory=dict)
    context_encoder: dict = Field(default_factory=dict)


class InvalidLoadType(Exception):
    pass


class HFSentenceTransformerRetriever(BaseRetriever):
    model_config = ConfigDict(protected_namespaces=("pydantic_model_",))
    model_name: str | None = Field(
        description="Name of HuggingFace SentenceTransformer model.",
        default=None,
    )
    query_model_name: str | None = Field(
        description="Name of HuggingFace SentenceTransformer model used for encoding queries.",
        default=None,
    )
    context_model_name: str | None = Field(
        description="Name of HuggingFace SentenceTransformer model used for encoding context.",
        default=None,
    )
    load_model_kwargs: LoadKwargs = Field(
        description="Optional kwargs dict for loading models from HF. Defaults to None.",
        default_factory=LoadKwargs,
    )
    _encoder: Optional["SentenceTransformer"] = PrivateAttr(default=None)
    _query_encoder: Optional["SentenceTransformer"] = PrivateAttr(default=None)
    _context_encoder: Optional["SentenceTransformer"] = PrivateAttr(
        default=None
    )

    def __init__(
        self,
        model_name: str | None = None,
        query_model_name: str | None = None,
        context_model_name: str | None = None,
        load_model_kwargs: LoadKwargs | dict | None = None,
        load_model_at_init: bool = True,
    ):
        if not _has_huggingface:
            msg = (
                f"`{self.__class__.__name__}` requires `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
            raise MissingExtraError(msg)

        if isinstance(load_model_kwargs, dict):
            # use same dict for all
            load_model_kwargs = LoadKwargs(
                encoder=load_model_kwargs,
                query_encoder=load_model_kwargs,
                context_encoder=load_model_kwargs,
            )

        load_model_kwargs = (
            load_model_kwargs if load_model_kwargs else LoadKwargs()
        )

        super().__init__(
            model_name=model_name,
            query_model_name=query_model_name,
            context_model_name=context_model_name,
            load_model_kwargs=load_model_kwargs,
        )
        if load_model_at_init:
            if model_name:
                self._encoder = self._load_model_from_hf(load_type="encoder")
            else:
                self._query_encoder = self._load_model_from_hf(
                    load_type="query_encoder"
                )
                self._context_encoder = self._load_model_from_hf(
                    load_type="context_encoder"
                )

    def _load_model_from_hf(
        self,
        load_type: Literal["encoder", "query_encoder", "context_encoder"],
        **kwargs: Any,
    ) -> "SentenceTransformer":
        if load_type == "encoder":
            load_kwargs = self.load_model_kwargs.encoder
            load_kwargs.update(kwargs)
            return SentenceTransformer(self.model_name, **load_kwargs)
        elif load_type == "context_encoder":
            load_kwargs = self.load_model_kwargs.context_encoder
            load_kwargs.update(kwargs)
            return SentenceTransformer(self.context_model_name, **load_kwargs)
        elif load_type == "query_encoder":
            load_kwargs = self.load_model_kwargs.query_encoder
            load_kwargs.update(kwargs)
            return SentenceTransformer(self.query_model_name, **load_kwargs)
        else:
            raise InvalidLoadType("Invalid `load_type` supplied.")

    def encode_context(
        self, context: str | list[str], **kwargs: Any
    ) -> torch.Tensor:
        # validation guarantees one of these is not None
        encoder = self.encoder if self.encoder else self.context_encoder
        encoder = cast(SentenceTransformer, encoder)

        return encoder.encode(context)

    def encode_query(
        self, query: str | list[str], **kwargs: Any
    ) -> torch.Tensor:
        # validation guarantees one of these is not None
        encoder = self.encoder if self.encoder else self.query_encoder
        encoder = cast(SentenceTransformer, encoder)

        return encoder.encode(query)

    @property
    def encoder(self) -> Optional["SentenceTransformer"]:
        if self.model_name and self._encoder is None:
            self._encoder = self._load_model_from_hf(load_type="encoder")
        return self._encoder

    @property
    def query_encoder(self) -> Optional["SentenceTransformer"]:
        if self.query_model_name and self._query_encoder is None:
            self._query_encoder = self._load_model_from_hf(
                load_type="query_encoder"
            )
        return self._query_encoder

    @property
    def context_encoder(self) -> Optional["SentenceTransformer"]:
        if self.context_model_name and self._context_encoder is None:
            self._context_encoder = self._load_model_from_hf(
                load_type="context_encoder"
            )
        return self._context_encoder



import pytest
from torch.utils.data import DataLoader, Dataset

from fed_rag.base.trainer import BaseTrainer
from fed_rag.exceptions import InconsistentDatasetError
from fed_rag.trainers.pytorch import PyTorchTrainerProtocol, TrainingArgs
from fed_rag.types.rag_system import RAGSystem

from .conftest import TestGeneratorTrainer, TestRetrieverTrainer


def test_pt_trainer_init(
    mock_rag_system: RAGSystem,
    train_dataset: Dataset,
    train_dataloader: DataLoader,
) -> None:
    # arrange
    training_args = TrainingArgs()
    trainer = TestGeneratorTrainer(
        rag_system=mock_rag_system,
        train_dataloader=train_dataloader,
        training_arguments=training_args,
    )

    assert trainer.rag_system == mock_rag_system
    assert trainer.model == mock_rag_system.generator.model
    assert trainer.train_dataset == train_dataset
    assert trainer.train_dataloader == train_dataloader
    assert trainer.training_arguments == training_args
    assert isinstance(trainer, PyTorchTrainerProtocol)
    assert isinstance(trainer, BaseTrainer)


def test_pt_retriever_trainer(
    mock_rag_system: RAGSystem,
    train_dataset: Dataset,
    train_dataloader: DataLoader,
) -> None:
    # arrange
    trainer = TestRetrieverTrainer(
        rag_system=mock_rag_system,
        train_dataloader=train_dataloader,
    )

    assert trainer.rag_system == mock_rag_system
    assert trainer.model == mock_rag_system.retriever.encoder
    assert trainer.train_dataset == train_dataset
    assert trainer.train_dataloader == train_dataloader
    assert trainer.training_arguments is None
    assert isinstance(trainer, PyTorchTrainerProtocol)
    assert isinstance(trainer, BaseTrainer)


def test_pt_trainer_init_raises_inconsistent_dataset_error(
    mock_rag_system: RAGSystem,
    another_train_dataset: Dataset,
    train_dataloader: DataLoader,
) -> None:
    msg = (
        "Inconsistent datasets detected between supplied `train_dataset` and that "
        "associated with the `train_dataloader`. These two datasets must be the same."
    )

    with pytest.raises(InconsistentDatasetError, match=msg):
        # arrange
        TestGeneratorTrainer(
            rag_system=mock_rag_system,
            train_dataset=another_train_dataset,
            train_dataloader=train_dataloader,
        )



from typing import Any

import pytest
import torch
from peft import LoraConfig, PeftModel, get_peft_model
from tokenizers import Tokenizer, models
from transformers import (
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)

from fed_rag.base.generator import BaseGenerator
from fed_rag.base.tokenizer import BaseTokenizer


class _TestHFConfig(PretrainedConfig):
    model_type = "testmodel"

    def __init__(self, num_hidden: int = 42, **kwargs: Any):
        super().__init__(**kwargs)
        self.num_hidden = num_hidden


class _TestHFPretrainedModel(PreTrainedModel):
    config_class = _TestHFConfig

    def __init__(self, config: _TestHFConfig):
        super().__init__(config)
        self.config = config
        self.model = torch.nn.Linear(3, self.config.num_hidden)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.model(input)


@pytest.fixture
def dummy_tokenizer() -> PreTrainedTokenizer:
    tokenizer = Tokenizer(models.WordPiece(unk_token="[UNK]"))
    return PreTrainedTokenizerFast(
        tokenizer_object=tokenizer,
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]",
    )


@pytest.fixture
def dummy_pretrained_model_and_tokenizer(
    dummy_tokenizer: PreTrainedTokenizer,
) -> tuple[PreTrainedModel, PreTrainedTokenizer]:
    return _TestHFPretrainedModel(_TestHFConfig()), dummy_tokenizer


@pytest.fixture
def dummy_peft_model_and_tokenizer(
    dummy_tokenizer: PreTrainedTokenizer,
) -> tuple[PeftModel, PreTrainedTokenizer]:
    class MLP(torch.nn.Module):
        """Taken from Peft's documentation, specifially 'Custom Models'.

        https://huggingface.co/docs/peft/main/en/developer_guides/custom_models
        """

        def __init__(self, num_units_hidden: int = 40):
            super().__init__()
            self.seq = torch.nn.Sequential(
                torch.nn.Linear(20, num_units_hidden),
                torch.nn.ReLU(),
                torch.nn.Linear(num_units_hidden, num_units_hidden),
                torch.nn.ReLU(),
                torch.nn.Linear(num_units_hidden, 2),
                torch.nn.LogSoftmax(dim=-1),
            )

        def forward(self, X: torch.Tensor) -> torch.Tensor:
            return self.seq(X)

    config = LoraConfig(
        target_modules=["seq.0", "seq.2"],
        r=4,
    )

    base_model = MLP()
    return get_peft_model(base_model, config), dummy_tokenizer


class MockTokenizer(BaseTokenizer):
    def encode(self, input: str, **kwargs: Any) -> list[int]:
        return [0, 1, 2]

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return "mock decoded sentence"

    @property
    def unwrapped(self) -> None:
        return None


@pytest.fixture()
def mock_tokenizer() -> BaseTokenizer:
    return MockTokenizer()


class MockGenerator(BaseGenerator):
    _model = torch.nn.Linear(2, 1)
    _tokenizer = MockTokenizer()

    def generate(self, input: str) -> str:
        return f"mock output from '{input}'."

    def compute_target_sequence_proba(self, prompt: str, target: str) -> float:
        return 0.42

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @property
    def tokenizer(self) -> MockTokenizer:
        return self._tokenizer


@pytest.fixture
def mock_generator() -> BaseGenerator:
    return MockGenerator()



from fed_rag.types.rag_system import RAGSystem

from .conftest import MockDataCollator


def test_init(mock_rag_system: RAGSystem) -> None:
    collator = MockDataCollator(rag_system=mock_rag_system)

    assert collator.rag_system == mock_rag_system


def test_collate(mock_rag_system: RAGSystem) -> None:
    collator = MockDataCollator(rag_system=mock_rag_system)

    # act
    res = collator(features={"feat": ["mock_input"]})

    assert collator.rag_system == mock_rag_system
    assert res == "collated!"



"""LM-Supervised Retriever Loss."""

from enum import Enum

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing_extensions import assert_never

from fed_rag.exceptions.loss import InvalidReductionParam


class ReductionMode(str, Enum):
    """Reduction mode enum."""

    MEAN = "mean"
    SUM = "sum"

    @classmethod
    def members_list(cls) -> list[str]:
        return [member for member in cls]


class LSRLoss(nn.Module):
    """PyTorch implementation of the LM-Supervised Retriever Loss.

    Given input context x and ground truth continuation y, computes KL divergence
    between retrieval likelihood P_R(d|x) and language model likelihood Q_LM(d|x,y),
    where d is the retrieved document.

    Source: Shi, Weijia, et al. "Replug: Retrieval-augmented black-box language models."
        arXiv preprint arXiv:2301.12652 (2023).
    Arxiv: https://arxiv.org/pdf/2301.12652
    """

    def __init__(self, reduction: ReductionMode = ReductionMode.MEAN):
        # This line is critical - it initializes all the Module machinery
        super(LSRLoss, self).__init__()

        if reduction not in ReductionMode.members_list():
            msg = (
                f"Invalid reduction {reduction}. "
                f"Valid reductions are: {', '.join(ReductionMode.members_list())}"
            )
            raise InvalidReductionParam(msg)

        self.reduction = reduction

    def forward(
        self, retrieval_scores: torch.Tensor, lm_scores: torch.Tensor
    ) -> torch.Tensor:
        retrieval_log_probs = F.log_softmax(retrieval_scores, dim=1)
        lm_probs = F.softmax(lm_scores, dim=1)
        kl_div = F.kl_div(retrieval_log_probs, lm_probs, reduction="none").sum(
            dim=-1
        )

        match self.reduction:
            case ReductionMode.MEAN:
                return kl_div.mean()
            case ReductionMode.SUM:
                return kl_div.sum()
            case _:  # pragma: no cover
                assert_never(self.reduction)  # pragma: no cover



from inspect import Parameter
from typing import cast


def get_type_name(t: Parameter) -> str | None:
    if isinstance(t.annotation, str):
        type_name = t.annotation
    else:
        type_name = getattr(
            t.annotation, "__name__", None
        )  # type:ignore [assignment]
    type_name = cast(str | None, type_name)  # type:ignore [assignment]
    return type_name



"""Knowledge Node"""

import json
import uuid
from enum import Enum
from typing import Any, TypedDict, cast

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_serializer,
    field_validator,
)


class NodeContent(TypedDict):
    text_content: str | None
    image_content: bytes | None


class NodeType(str, Enum):
    TEXT = "text"
    IMAGE = "image"
    MULTIMODAL = "multimodal"


class KnowledgeNode(BaseModel):
    model_config = ConfigDict(
        # ensures that validation is performed for defaulted None values
        validate_default=True
    )
    node_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    embedding: list[float] = Field(
        description="Encoded representation of node. If multimodal type, then this is shared embedding between image and text."
    )
    node_type: NodeType = Field(description="Type of node.")
    text_content: str | None = Field(
        description="Text content. Used for TEXT and potentially MULTIMODAL node types.",
        default=None,
    )
    image_content: bytes | None = Field(
        description="Image content as binary data (decoded from base64)",
        default=None,
    )
    metadata: dict = Field(
        description="Metadata for node.", default_factory=dict
    )

    # validators
    @field_validator("text_content", mode="before")
    @classmethod
    def validate_text_content(
        cls, value: str | None, info: ValidationInfo
    ) -> str | None:
        node_type = info.data.get("node_type")
        node_type = cast(NodeType, node_type)
        if node_type == NodeType.TEXT and value is None:
            raise ValueError("NodeType == 'text', but text_content is None.")

        if node_type == NodeType.MULTIMODAL and value is None:
            raise ValueError(
                "NodeType == 'multimodal', but text_content is None."
            )

        return value

    @field_validator("image_content", mode="after")
    @classmethod
    def validate_image_content(
        cls, value: str | None, info: ValidationInfo
    ) -> str | None:
        node_type = info.data.get("node_type")
        node_type = cast(NodeType, node_type)
        if node_type == NodeType.IMAGE:
            if value is None:
                raise ValueError(
                    "NodeType == 'image', but image_content is None."
                )

        if node_type == NodeType.MULTIMODAL:
            if value is None:
                raise ValueError(
                    "NodeType == 'multimodal', but image_content is None."
                )

        return value

    def get_content(self) -> NodeContent:
        """Return dict of node content."""
        content: NodeContent = {
            "image_content": self.image_content,
            "text_content": self.text_content,
        }
        return content

    @field_serializer("metadata")
    def serialize_metadata(
        self, metadata: dict[Any, Any] | None
    ) -> str | None:
        """
        Custom serializer for the metadata field.

        Will serialize the metadata field into a json string.

        Args:
            metadata: Metadata dictionary to serialize.

        Returns:
            Serialized metadata as a json string.
        """
        if metadata:
            return json.dumps(metadata)
        return None

    @field_validator("metadata", mode="before")
    @classmethod
    def deserialize_metadata(
        cls, metadata: dict[Any, Any] | str | None
    ) -> dict[Any, Any] | None:
        """
        Custom validator for the metadata field.

        Will deserialize the metadata from a json string if it's a string.

        Args:
            metadata: Metadata to validate. If it is a json string, it will be deserialized into a dictionary.

        Returns:
            Validated metadata.
        """
        if isinstance(metadata, str):
            deserialized_metadata = json.loads(metadata)
            return cast(dict[Any, Any], deserialized_metadata)
        if metadata is None:
            return {}
        return metadata



"""Base Trainer"""

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict, PrivateAttr, model_validator

from fed_rag.types.rag_system import RAGSystem
from fed_rag.types.results import TestResult, TrainResult


class BaseTrainer(BaseModel, ABC):
    """Base Trainer Class."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    rag_system: RAGSystem
    train_dataset: Any
    _model = PrivateAttr()

    @abstractmethod
    def train(self) -> TrainResult:
        """Train loop."""

    @abstractmethod
    def evaluate(self) -> TestResult:
        """Evaluation"""

    @abstractmethod
    def _get_model_from_rag_system(self) -> Any:
        """Get the model from the RAG system."""

    @model_validator(mode="after")
    def set_model(self) -> "BaseTrainer":
        self._model = self._get_model_from_rag_system()
        return self

    @property
    def model(self) -> Any:
        """Return the model to be trained."""
        return self._model

    @model.setter
    def model(self, v: Any) -> None:
        """Set the model to be trained."""
        self._model = v


class BaseRetrieverTrainer(BaseTrainer, ABC):
    """Base Retriever Trainer Class."""

    def _get_model_from_rag_system(self) -> Any:
        if self.rag_system.retriever.encoder:
            return self.rag_system.retriever.encoder
        else:
            return (
                self.rag_system.retriever.query_encoder
            )  # only update query encoder


class BaseGeneratorTrainer(BaseTrainer, ABC):
    """Base Retriever Trainer Class."""

    def _get_model_from_rag_system(self) -> Any:
        return self.rag_system.generator.model



"""LlamaIndex Bridge"""

from typing import TYPE_CHECKING

from fed_rag.base.bridge import BaseBridgeMixin
from fed_rag.bridges.llamaindex._version import __version__

if TYPE_CHECKING:  # pragma: no cover
    from llama_index.core.indices.managed.base import BaseManagedIndex

    from fed_rag.types.rag_system import RAGSystem  # avoids circular import


class LlamaIndexBridgeMixin(BaseBridgeMixin):
    """LlamaIndex Bridge.

    This mixin adds LlamaIndex conversion capabilities to RAGSystem.
    When mixed with a RAGSystem, it allows direct conversion to
    LlamaIndex's BaseManagedIndex through the to_llamaindex() method.
    """

    _bridge_version = __version__
    _bridge_extra = "llama-index"
    _framework = "llama-index"
    _compatible_versions = ["0.12.35"]
    _method_name = "to_llamaindex"

    def to_llamaindex(self: "RAGSystem") -> "BaseManagedIndex":
        """Converts the RAGSystem to a ~llamaindex.core.BaseManagedIndex."""
        self._validate_framework_installed()

        from fed_rag.bridges.llamaindex._managed_index import (
            FedRAGManagedIndex,
        )

        return FedRAGManagedIndex(rag_system=self)



import re
import sys
from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset
from pytest import MonkeyPatch
from transformers import Trainer
from transformers.trainer_utils import TrainOutput

from fed_rag.exceptions import FedRAGError, MissingExtraError
from fed_rag.trainers.huggingface.ralt import HuggingFaceTrainerForRALT
from fed_rag.types.rag_system import RAGSystem


def test_init(
    hf_rag_system: RAGSystem, train_dataset: Dataset, monkeypatch: MonkeyPatch
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForRALT(
        rag_system=hf_rag_system,
        train_dataset=train_dataset,
    )

    assert trainer.train_dataset == train_dataset
    assert trainer.model == hf_rag_system.generator.model
    assert trainer.rag_system == hf_rag_system
    assert trainer.training_arguments.remove_unused_columns is False
    assert isinstance(trainer.hf_trainer_obj, Trainer)


def test_init_raises_invalid_rag_system_error(
    mock_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    with pytest.raises(FedRAGError):
        HuggingFaceTrainerForRALT(
            rag_system=mock_rag_system,
            train_dataset=train_dataset,
        )


def test_huggingface_extra_missing(
    train_dataset: Dataset, hf_rag_system: RAGSystem, monkeypatch: MonkeyPatch
) -> None:
    modules = {
        "transformers": None,
    }
    modules_to_import = [
        "fed_rag.trainers.huggingface.mixin",
        "fed_rag.trainers.huggingface.ralt",
    ]
    original_modules = [sys.modules.pop(m, None) for m in modules_to_import]

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceTrainerForRALT` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.trainers.huggingface.ralt import (
                HuggingFaceTrainerForRALT,
            )

            HuggingFaceTrainerForRALT(
                rag_system=hf_rag_system,
                train_dataset=train_dataset,
            )

    # restore module so to not affect other tests
    for ix, original_module in enumerate(original_modules):
        if original_module:
            sys.modules[modules_to_import[ix]] = original_module


@patch.object(HuggingFaceTrainerForRALT, "hf_trainer_obj")
def test_train(
    mock_hf_trainer: MagicMock,
    hf_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForRALT(
        rag_system=hf_rag_system,
        train_dataset=train_dataset,
    )
    mock_hf_trainer.train.return_value = TrainOutput(
        global_step=42, training_loss=0.42, metrics={}
    )

    out = trainer.train()

    mock_hf_trainer.train.assert_called_once()
    assert out.loss == 0.42


def test_evaluate(
    hf_rag_system: RAGSystem,
    train_dataset: Dataset,
    monkeypatch: MonkeyPatch,
) -> None:
    # skip validation of rag system
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    trainer = HuggingFaceTrainerForRALT(
        train_dataset=train_dataset,
        rag_system=hf_rag_system,
    )

    with pytest.raises(NotImplementedError):
        trainer.evaluate()



"""TesterResult"""

from typing import Any

from pydantic import BaseModel, Field


class TrainResult(BaseModel):
    loss: float


class TestResult(BaseModel):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )
    loss: float
    metrics: dict[str, Any] = Field(
        description="Additional metrics computed on test set.",
        default_factory=dict,
    )



import re
import sys
from unittest.mock import MagicMock, _Call, patch

import pytest
import torch
from pytest import MonkeyPatch
from torch.testing import assert_close

from fed_rag.data_collators.huggingface import DataCollatorForLSR
from fed_rag.exceptions import MissingExtraError
from fed_rag.exceptions.core import FedRAGError
from fed_rag.generators.huggingface import HFPeftModelGenerator
from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
    HFSentenceTransformerRetriever,
)
from fed_rag.types.knowledge_node import KnowledgeNode
from fed_rag.types.rag_system import RAGSystem, SourceNode


def test_huggingface_extra_missing(mock_rag_system: RAGSystem) -> None:
    """Test extra is not installed."""

    modules = {
        "sentence_transformers": None,
        "sentence_transformers.data_collator": None,
    }
    module_to_import = "fed_rag.data_collators.huggingface.lsr"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`DataCollatorForLSR` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.data_collators.huggingface.lsr import (
                DataCollatorForLSR,
            )

            DataCollatorForLSR(
                rag_system=mock_rag_system,
                prompt_template="{query} and {context}",
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module


def test_huggingface_extra_missing_from_parent(
    mock_rag_system: RAGSystem,
) -> None:
    """Test extra is not installed."""

    modules = {
        "transformers": None,
        "transformers.data": None,
        "transformers.data.data_collator": None,
    }
    module_to_import = "fed_rag.data_collators.huggingface"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`fed_rag.data_collators.huggingface` requires `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.data_collators.huggingface import DataCollatorForLSR

            DataCollatorForLSR(
                rag_system=mock_rag_system,
                prompt_template="{query} and {context}",
            )

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module


def test_invalid_rag_system_due_to_generators(
    mock_rag_system: RAGSystem,
) -> None:
    with pytest.raises(
        FedRAGError,
        match="Generator must be HFPretrainedModelGenerator or HFPeftModelGenerator",
    ):
        from fed_rag.data_collators.huggingface import DataCollatorForLSR

        DataCollatorForLSR(rag_system=mock_rag_system, prompt_template="")


def test_invalid_rag_system_due_to_retriever(
    mock_rag_system: RAGSystem,
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    mock_rag_system.generator = generator

    with pytest.raises(
        FedRAGError,
        match="Retriever must be a HFSentenceTransformerRetriever",
    ):
        from fed_rag.data_collators.huggingface import DataCollatorForLSR

        DataCollatorForLSR(rag_system=mock_rag_system, prompt_template="")


def test_invalid_return_tensors(
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")
    return_tensors = "af"
    with pytest.raises(
        FedRAGError,
        match=f"Framework '{return_tensors}' not recognized!",
    ):
        collator = DataCollatorForLSR(
            rag_system=mock_rag_system, prompt_template=""
        )
        collator([], return_tensors)


def test_init(
    mock_rag_system: RAGSystem,
) -> None:
    generator = HFPeftModelGenerator(
        model_name="fake_name",
        base_model_name="fake_base_name",
        load_model_at_init=False,
    )
    retriever = HFSentenceTransformerRetriever(
        model_name="fake_name", load_model_at_init=False
    )
    mock_rag_system.generator = generator
    mock_rag_system.retriever = retriever

    collator = DataCollatorForLSR(
        rag_system=mock_rag_system, prompt_template="{query} and {context}"
    )

    assert collator.rag_system == mock_rag_system
    assert collator.prompt_template == "{query} and {context}"
    assert collator.default_return_tensors == "pt"


@patch.object(RAGSystem, "retrieve")
def test_lsr_collator_with_mocks(
    mock_retrieve: MagicMock,
    mock_rag_system: RAGSystem,
    monkeypatch: MonkeyPatch,
) -> None:
    # Set environment variable for the duration of this test only
    monkeypatch.setenv("FEDRAG_SKIP_VALIDATION", "1")

    rag_system = RAGSystem(
        generator=mock_rag_system.generator,
        retriever=mock_rag_system.retriever,
        knowledge_store=mock_rag_system.knowledge_store,
        rag_config=mock_rag_system.rag_config,
    )

    # use mocks
    mock_generator = MagicMock()
    mock_generator.compute_target_sequence_proba.side_effect = [
        torch.tensor(0.01),
        torch.tensor(0.02),
        torch.tensor(0.03),
    ]
    rag_system.generator = mock_generator

    mock_retrieve.return_value = [
        SourceNode(
            score=0.1,
            node=KnowledgeNode(
                embedding=[0.1], node_type="text", text_content="node 1"
            ),
        ),
        SourceNode(
            score=0.2,
            node=KnowledgeNode(
                embedding=[0.2], node_type="text", text_content="node 2"
            ),
        ),
        SourceNode(
            score=0.3,
            node=KnowledgeNode(
                embedding=[0.3], node_type="text", text_content="node 3"
            ),
        ),
    ]

    collator = DataCollatorForLSR(
        rag_system=rag_system, prompt_template="{query} {context}"
    )

    # act
    features = [{"query": "mock query", "response": "mock response"}]
    batch = collator(features)

    assert_close(
        batch["retrieval_scores"], torch.tensor([0.1, 0.2, 0.3]).unsqueeze(0)
    )
    assert_close(
        batch["lm_scores"], torch.tensor([0.01, 0.02, 0.03]).unsqueeze(0)
    )
    mock_retrieve.assert_called_once_with("mock query")
    mock_generator.compute_target_sequence_proba.assert_has_calls(
        [
            _Call(
                (
                    (),
                    {
                        "prompt": "mock query node 1",
                        "target": "\n<response>\nmock response\n</response>\n",
                    },
                )
            ),
            _Call(
                (
                    (),
                    {
                        "prompt": "mock query node 2",
                        "target": "\n<response>\nmock response\n</response>\n",
                    },
                )
            ),
            _Call(
                (
                    (),
                    {
                        "prompt": "mock query node 3",
                        "target": "\n<response>\nmock response\n</response>\n",
                    },
                )
            ),
        ]
    )



"""Base Generator"""

from abc import ABC, abstractmethod

import torch
from pydantic import BaseModel, ConfigDict

from fed_rag.base.tokenizer import BaseTokenizer


class BaseGenerator(BaseModel, ABC):
    """Base Generator Class."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def generate(self, query: str, context: str, **kwargs: dict) -> str:
        """Generate an output from a given query and context."""

    @property
    @abstractmethod
    def model(self) -> torch.nn.Module:
        """Model associated with this generator."""

    @property
    @abstractmethod
    def tokenizer(self) -> BaseTokenizer:
        """Tokenizer associated with this generator."""

    @abstractmethod
    def compute_target_sequence_proba(
        self, prompt: str, target: str
    ) -> torch.Tensor:
        """Compute P(target | prompt).

        NOTE: this is used in LM Supervised Retriever fine-tuning.
        """



from unittest.mock import MagicMock, patch

import pytest
from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.schema import MediaResource
from llama_index.core.schema import Node as LlamaNode
from llama_index.core.schema import QueryBundle

from fed_rag.bridges.llamaindex._managed_index import (
    FedRAGManagedIndex,
    convert_llama_index_node_to_knowledge_node,
    convert_source_node_to_llama_index_node_with_score,
)
from fed_rag.bridges.llamaindex.bridge import LlamaIndexBridgeMixin
from fed_rag.exceptions import BridgeError
from fed_rag.types.rag_system import KnowledgeNode, RAGSystem, SourceNode


def test_rag_system_bridges(mock_rag_system: RAGSystem) -> None:
    metadata = LlamaIndexBridgeMixin.get_bridge_metadata()

    assert "llama-index" in mock_rag_system.bridges
    assert mock_rag_system.bridges[metadata["framework"]] == metadata
    assert LlamaIndexBridgeMixin._bridge_extra == "llama-index"


@patch("fed_rag.bridges.llamaindex._managed_index.FedRAGManagedIndex")
def test_rag_system_conversion_method(
    mock_managed_index_class: MagicMock, mock_rag_system: RAGSystem
) -> None:
    metadata = LlamaIndexBridgeMixin.get_bridge_metadata()

    conversion_method = getattr(mock_rag_system, metadata["method_name"])
    conversion_method()

    mock_managed_index_class.assert_called_once_with(
        rag_system=mock_rag_system
    )


# test node converters
def test_convert_llama_node_to_knowledge_node() -> None:
    llama_node = LlamaNode(
        id_="1",
        embedding=[1, 1, 1],
        text_resource=MediaResource(text="mock text"),
    )
    fed_rag_node = convert_llama_index_node_to_knowledge_node(llama_node)

    assert fed_rag_node.node_id == "1"
    assert fed_rag_node.embedding == [1, 1, 1]
    assert fed_rag_node.node_type == "text"
    assert fed_rag_node.text_content == "mock text"


def test_convert_llama_node_to_knowledge_node_raises_error_missing_embedding() -> (
    None
):
    llama_node = LlamaNode(
        id_="1", text_resource=MediaResource(text="mock text")
    )

    with pytest.raises(
        BridgeError,
        match="Failed to convert ~llama_index.Node: embedding attribute is None.",
    ):
        convert_llama_index_node_to_knowledge_node(llama_node)


def test_convert_llama_node_to_knowledge_node_raises_error_text_resource_is_none() -> (
    None
):
    llama_node = LlamaNode(id_="1", embedding=[1, 1, 1])

    with pytest.raises(
        BridgeError,
        match="Failed to convert ~llama_index.Node: text_resource attribute is None.",
    ):
        convert_llama_index_node_to_knowledge_node(llama_node)


def test_convert_source_node_to_llama_node_with_score() -> None:
    source_node = SourceNode(
        score=0.42,
        node=KnowledgeNode(
            node_id="1",
            embedding=[1, 1, 1],
            node_type="text",
            text_content="mock text",
        ),
    )

    llama_node_with_score = convert_source_node_to_llama_index_node_with_score(
        source_node
    )

    assert llama_node_with_score.score == 0.42
    assert isinstance(llama_node_with_score.node, LlamaNode)
    assert llama_node_with_score.node.id_ == "1"
    assert llama_node_with_score.node.text_resource.text == "mock text"


# test FedRAGManagedIndex
def test_fedrag_managed_index_init(mock_rag_system: RAGSystem) -> None:
    index = FedRAGManagedIndex(rag_system=mock_rag_system)

    assert index._rag_system == mock_rag_system


def test_fedrag_managed_index_init_raises_error_nodes_passed(
    mock_rag_system: RAGSystem,
) -> None:
    with pytest.raises(
        BridgeError,
        match="FedRAGManagedIndex does not support nodes on initialization.",
    ):
        FedRAGManagedIndex(rag_system=mock_rag_system, nodes=[LlamaNode()])


def test_fedrag_managed_index_as_retriever(mock_rag_system: RAGSystem) -> None:
    index = FedRAGManagedIndex(rag_system=mock_rag_system)
    retriever = index.as_retriever()

    assert isinstance(retriever, FedRAGManagedIndex.FedRAGRetriever)
    assert retriever._rag_system == mock_rag_system


def test_fedrag_managed_index_as_retriever_retrieve_method(
    mock_rag_system: RAGSystem,
) -> None:
    mock_rag_system.knowledge_store.load_node(
        KnowledgeNode(
            node_id="1",
            node_type="text",
            text_content="mock node",
            embedding=[1, 1, 1],
        )
    )
    index = FedRAGManagedIndex(rag_system=mock_rag_system)
    retriever = index.as_retriever()

    # act
    res = retriever._retrieve(query_bundle=QueryBundle(query_str="mock query"))

    assert len(res) == 1
    assert res[0].node_id == "1"
    assert res[0].node.text_resource.text == "mock node"


@patch("llama_index.core.indices.base.resolve_llm")
@patch(
    "fed_rag.bridges.llamaindex._managed_index.FedRAGManagedIndex.FedRAGLLM"
)
def test_fedrag_managed_index_as_query_engine_mocked(
    mock_fedrag_llm_class: MagicMock,
    mock_resolve_llm: MagicMock,
    mock_rag_system: RAGSystem,
) -> None:
    index = FedRAGManagedIndex(rag_system=mock_rag_system)
    query_engine = index.as_query_engine()

    assert isinstance(query_engine, BaseQueryEngine)
    mock_fedrag_llm_class.assert_called_once_with(rag_system=mock_rag_system)
    mock_resolve_llm.assert_called_once()


def test_fedrag_managed_index_as_query_engine(
    mock_rag_system: RAGSystem,
) -> None:
    index = FedRAGManagedIndex(rag_system=mock_rag_system)
    query_engine = index.as_query_engine()

    assert isinstance(query_engine, BaseQueryEngine)


def test_fedrag_managed_index_delete_node(
    mock_rag_system: RAGSystem,
) -> None:
    mock_rag_system.knowledge_store.load_node(
        KnowledgeNode(
            node_id="1",
            node_type="text",
            text_content="mock node",
            embedding=[1, 1, 1],
        )
    )
    index = FedRAGManagedIndex(rag_system=mock_rag_system)

    # act
    index._delete_node(node_id="1")

    # assert
    assert index._rag_system.knowledge_store.count == 0


def test_fedrag_managed_index_insert(
    mock_rag_system: RAGSystem,
) -> None:
    llama_nodes = [
        LlamaNode(
            id_="1",
            embedding=[1, 1, 1],
            text_resource=MediaResource(text="node 1"),
        ),
        LlamaNode(
            id_="2",
            embedding=[2, 2, 2],
            text_resource=MediaResource(text="node 2"),
        ),
    ]
    index = FedRAGManagedIndex(rag_system=mock_rag_system)

    # act
    index._insert(nodes=llama_nodes)

    # assert
    assert index._rag_system.knowledge_store.count == 2


def test_fedrag_llm_complete(mock_rag_system: RAGSystem) -> None:
    llm = FedRAGManagedIndex.FedRAGLLM(mock_rag_system)

    response = llm.complete("mock prompt")

    assert response.text == "mock output from 'mock prompt' and ''."


def test_fedrag_llm_stream_complete_raises_error(
    mock_rag_system: RAGSystem,
) -> None:
    llm = FedRAGManagedIndex.FedRAGLLM(mock_rag_system)

    with pytest.raises(
        NotImplementedError,
        match="stream_complete is not implemented for FedRAGLLM.",
    ):
        llm.stream_complete(prompt="mock prompt")


## test methods with no implementation
def test_fedrag_managed_index_raises_not_implemented_error(
    mock_rag_system: RAGSystem,
) -> None:
    from llama_index.core.schema import Document

    index = FedRAGManagedIndex(rag_system=mock_rag_system)

    with pytest.raises(
        NotImplementedError,
        match="update_ref_doc not implemented for `FedRAGManagedIndex`.",
    ):
        index.update_ref_doc(document=Document())

    with pytest.raises(
        NotImplementedError,
        match="delete_ref_doc not implemented for `FedRAGManagedIndex`.",
    ):
        index.delete_ref_doc(ref_doc_id="1")



from fed_rag.exceptions import FedRAGError
from fed_rag.types.rag_system import RAGSystem


def _validate_rag_system(rag_system: RAGSystem) -> None:
    # Skip validation if environment variable is set
    import os

    if os.environ.get("FEDRAG_SKIP_VALIDATION") == "1":
        return

    from fed_rag.generators.huggingface import (
        HFPeftModelGenerator,
        HFPretrainedModelGenerator,
    )
    from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
        HFSentenceTransformerRetriever,
    )

    if not isinstance(
        rag_system.generator, HFPretrainedModelGenerator
    ) and not isinstance(rag_system.generator, HFPeftModelGenerator):
        raise FedRAGError(
            "Generator must be HFPretrainedModelGenerator or HFPeftModelGenerator"
        )

    if not isinstance(rag_system.retriever, HFSentenceTransformerRetriever):
        raise FedRAGError("Retriever must be a HFSentenceTransformerRetriever")



from typing import Any

import numpy as np
import pytest
from torch.utils.data import DataLoader, Dataset

from fed_rag.base.trainer import BaseGeneratorTrainer, BaseRetrieverTrainer
from fed_rag.trainers.pytorch.mixin import PyTorchTrainerMixin
from fed_rag.types.results import TestResult, TrainResult


class TestRetrieverTrainer(PyTorchTrainerMixin, BaseRetrieverTrainer):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )

    def train(self) -> TrainResult:
        return TrainResult(loss=0.42)

    def evaluate(self) -> TestResult:
        return TestResult(loss=0.42)


class TestGeneratorTrainer(PyTorchTrainerMixin, BaseGeneratorTrainer):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )

    def train(self) -> TrainResult:
        return TrainResult(loss=0.42)

    def evaluate(self) -> TestResult:
        return TestResult(loss=0.42)


class _TestDataset(Dataset):
    def __init__(self, size: int) -> None:
        self.features = np.random.rand(size, 2)
        self.labels = np.random.choice(2, size=size)

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, index: int) -> tuple[np.ndarray, Any]:
        return self.features[index], self.labels[index]


@pytest.fixture()
def train_dataset() -> Dataset:
    return _TestDataset(size=10)


@pytest.fixture()
def another_train_dataset() -> Dataset:
    return _TestDataset(size=10)


@pytest.fixture()
def train_dataloader(train_dataset: Dataset) -> DataLoader:
    return DataLoader(train_dataset, batch_size=2, shuffle=True)



import re
import sys
from unittest.mock import patch

import pytest
import torch
from datasets import Dataset

from fed_rag.exceptions import MissingExtraError
from fed_rag.utils.data.finetuning_datasets.huggingface import (
    HuggingFaceRAGFinetuningDataset,
)


def test_hf_rag_ft_dataset_init(
    input_and_target_ids: tuple[torch.Tensor, torch.Tensor],
) -> None:
    input_ids, target_ids = input_and_target_ids
    rag_ft_dataset = HuggingFaceRAGFinetuningDataset.from_inputs(
        input_ids=input_ids,
        target_ids=target_ids,
        attention_mask=[None] * len(target_ids),
    )

    assert len(rag_ft_dataset) == len(input_ids)
    assert isinstance(rag_ft_dataset, Dataset)
    assert rag_ft_dataset["input_ids"] == [t.tolist() for t in input_ids]
    assert rag_ft_dataset["target_ids"] == [t.tolist() for t in target_ids]
    assert rag_ft_dataset["attention_mask"] == [None, None, None]


def test_hf_rag_ft_dataset_missing_extra_raises_error(
    input_and_target_ids: tuple[torch.Tensor, torch.Tensor],
) -> None:
    input_ids, target_ids = input_and_target_ids
    modules = {"datasets": None}
    module_to_import = "fed_rag.utils.data.finetuning_datasets.huggingface"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceRAGFinetuningDataset` requires the `huggingface` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface]`."
        )

        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.utils.data.finetuning_datasets.huggingface import (
                HuggingFaceRAGFinetuningDataset,
            )

            HuggingFaceRAGFinetuningDataset.from_inputs(
                input_ids=input_ids, target_ids=target_ids
            )

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module



"""Qdrant utils module."""

from importlib.util import find_spec

from fed_rag.exceptions import MissingExtraError


def check_qdrant_installed() -> None:
    if find_spec("qdrant_client") is None:
        raise MissingExtraError(
            "Qdrant knowledge stores require the qdrant-client to be installed. "
            "To fix please run `pip install fed-rag[qdrant]`."
        )



from typing import Sequence
from unittest.mock import MagicMock

import pytest
import torch

from fed_rag.base.tokenizer import EncodeResult
from fed_rag.types.rag_system import KnowledgeNode, SourceNode
from fed_rag.utils.data import build_finetune_dataset
from fed_rag.utils.data.finetuning_datasets import PyTorchRAGFinetuningDataset
from fed_rag.utils.data.finetuning_datasets.huggingface import (
    HuggingFaceRAGFinetuningDataset,
)


@pytest.fixture()
def mock_examples() -> Sequence[dict]:
    return [
        {
            "query": f"fake query {ix}",
            "answer": f"fake answer {ix}",
            "context": f"fake context {ix}",
        }
        for ix in range(2)
    ]


@pytest.fixture()
def mock_source_nodes() -> list[list[SourceNode]]:
    return [
        [
            SourceNode(
                score=ix / 10,
                node=KnowledgeNode(
                    embedding=[ix, ix, ix],
                    node_type="text",
                    text_content=f"fake text context {ix}",
                ),
            )
            for ix in range(2)
        ],
        [
            SourceNode(
                score=ix / 10,
                node=KnowledgeNode(
                    embedding=[ix, ix, ix],
                    node_type="text",
                    text_content=f"fake text context {ix}",
                ),
            )
            for ix in range(2, 4)
        ],
    ]


def test_build_finetune_dataset_txt_return(
    mock_examples: Sequence[dict], mock_source_nodes: list[list[SourceNode]]
) -> None:
    # arrange
    mock_rag_system = MagicMock()
    mock_retrieve = MagicMock()
    mock_retrieve.side_effect = mock_source_nodes
    mock_tokenizer = MagicMock()
    mock_encode_return: EncodeResult = {
        "attention_mask": None,
        "input_ids": [1, 1, 1],
    }
    mock_tokenizer.encode.return_value = mock_encode_return
    mock_rag_system.retrieve = mock_retrieve
    mock_rag_system.generator.tokenizer = mock_tokenizer

    # act
    result = build_finetune_dataset(
        rag_system=mock_rag_system,
        examples=mock_examples,
        eos_token_id=42,
        return_dataset="txt",
    )

    # assert
    assert result == [
        "fake query 0 fake text context 0 fake answer 0",
        "fake query 0 fake text context 1 fake answer 0",
        "fake query 1 fake text context 2 fake answer 1",
        "fake query 1 fake text context 3 fake answer 1",
    ]


def test_build_finetune_dataset_pt_return(
    mock_examples: Sequence[dict], mock_source_nodes: list[list[SourceNode]]
) -> None:
    # arrange
    mock_rag_system = MagicMock()
    mock_retrieve = MagicMock()
    mock_retrieve.side_effect = mock_source_nodes
    mock_tokenizer = MagicMock()
    mock_encode_return: EncodeResult = {
        "attention_mask": None,
        "input_ids": [1, 1, 1],
    }
    mock_tokenizer.encode.return_value = mock_encode_return
    mock_rag_system.retrieve = mock_retrieve
    mock_rag_system.generator.tokenizer = mock_tokenizer

    # act
    result: PyTorchRAGFinetuningDataset = build_finetune_dataset(
        rag_system=mock_rag_system,
        examples=mock_examples,
        eos_token_id=42,
        return_dataset="pt",
    )

    # assert
    assert isinstance(result, PyTorchRAGFinetuningDataset)
    assert len(result) == 4
    assert isinstance(result.input_ids[0], torch.Tensor)
    assert isinstance(result.target_ids[0], torch.Tensor)


def test_build_finetune_dataset_hf_return(
    mock_examples: Sequence[dict], mock_source_nodes: list[list[SourceNode]]
) -> None:
    # arrange
    mock_rag_system = MagicMock()
    mock_retrieve = MagicMock()
    mock_retrieve.side_effect = mock_source_nodes
    mock_tokenizer = MagicMock()
    mock_encode_return: EncodeResult = {
        "attention_mask": None,
        "input_ids": [1, 1, 1],
    }
    mock_tokenizer.encode.return_value = mock_encode_return
    mock_rag_system.retrieve = mock_retrieve
    mock_rag_system.generator.tokenizer = mock_tokenizer

    # act
    result: HuggingFaceRAGFinetuningDataset = build_finetune_dataset(
        rag_system=mock_rag_system,
        examples=mock_examples,
        eos_token_id=42,
        return_dataset="hf",
    )

    # assert
    assert isinstance(result, HuggingFaceRAGFinetuningDataset)
    assert len(result) == 4
    assert result.column_names == ["input_ids", "target_ids", "attention_mask"]
    assert result[:2] == {
        "input_ids": [
            [1, 1, 1],
            [
                1,
                1,
                1,
            ],
        ],
        "target_ids": [[1, 1, 42], [1, 1, 42]],
        "attention_mask": [None, None],
    }


def test_build_finetune_dataset_invalid_return_raises_error(
    mock_examples: Sequence[dict], mock_source_nodes: list[list[SourceNode]]
) -> None:
    # arrange
    mock_rag_system = MagicMock()
    mock_retrieve = MagicMock()
    mock_retrieve.side_effect = mock_source_nodes
    mock_tokenizer = MagicMock()
    mock_encode_return: EncodeResult = {
        "attention_mask": None,
        "input_ids": [1, 1, 1],
    }
    mock_tokenizer.encode.return_value = mock_encode_return
    mock_rag_system.retrieve = mock_retrieve
    mock_rag_system.generator.tokenizer = mock_tokenizer

    with pytest.raises(ValueError, match="Invalid `return_type` specified."):
        build_finetune_dataset(
            rag_system=mock_rag_system,
            examples=mock_examples,
            eos_token_id=42,
            return_dataset="invalid_return",
        )



from .bridge import BridgeError, MissingSpecifiedConversionMethod
from .common import MissingExtraError
from .core import FedRAGError
from .data_collator import DataCollatorError
from .fl_tasks import (
    FLTaskError,
    MissingFLTaskConfig,
    MissingRequiredNetParam,
    NetTypeMismatch,
)
from .inspectors import (
    InspectorError,
    InspectorWarning,
    InvalidReturnType,
    MissingDataParam,
    MissingMultipleDataParams,
    MissingNetParam,
    MissingTesterSpec,
    MissingTrainerSpec,
    UnequalNetParamWarning,
)
from .knowledge_stores import (
    InvalidDistanceError,
    KnowledgeStoreError,
    KnowledgeStoreNotFoundError,
    KnowledgeStoreWarning,
    LoadNodeError,
)
from .trainer import (
    InconsistentDatasetError,
    InvalidDataCollatorError,
    InvalidLossError,
    MissingInputTensor,
    TrainerError,
)
from .trainer_manager import (
    InconsistentRAGSystems,
    RAGTrainerManagerError,
    UnspecifiedGeneratorTrainer,
    UnspecifiedRetrieverTrainer,
    UnsupportedTrainerMode,
)

__all__ = [
    # core
    "FedRAGError",
    # common
    "MissingExtraError",
    "DataCollatorError",
    # bridges"
    "BridgeError",
    "MissingSpecifiedConversionMethod",
    # fl_tasks
    "FLTaskError",
    "MissingFLTaskConfig",
    "MissingRequiredNetParam",
    "NetTypeMismatch",
    # inspectors
    "InspectorError",
    "InspectorWarning",
    "MissingNetParam",
    "MissingMultipleDataParams",
    "MissingDataParam",
    "MissingTrainerSpec",
    "MissingTesterSpec",
    "UnequalNetParamWarning",
    "InvalidReturnType",
    # knowledge stores
    "KnowledgeStoreError",
    "KnowledgeStoreWarning",
    "KnowledgeStoreNotFoundError",
    "InvalidDistanceError",
    "LoadNodeError",
    # rag trainer manager
    "RAGTrainerManagerError",
    "UnspecifiedGeneratorTrainer",
    "UnspecifiedRetrieverTrainer",
    "UnsupportedTrainerMode",
    "InconsistentRAGSystems",
    # trainer
    "TrainerError",
    "InvalidLossError",
    "MissingInputTensor",
    "InvalidDataCollatorError",
    "InconsistentDatasetError",
]



import pytest
import torch
from datasets import Dataset
from sentence_transformers import SentenceTransformer
from transformers import Trainer

from fed_rag.base.trainer import BaseGeneratorTrainer, BaseRetrieverTrainer
from fed_rag.trainers.huggingface.mixin import HuggingFaceTrainerMixin
from fed_rag.types.rag_system import RAGSystem
from fed_rag.types.results import TestResult, TrainResult


class TestHFRetrieverTrainer(HuggingFaceTrainerMixin, BaseRetrieverTrainer):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )

    def train(self) -> TrainResult:
        return TrainResult(loss=0.42)

    def evaluate(self) -> TestResult:
        return TestResult(loss=0.42)

    def hf_trainer_obj(self) -> Trainer:
        return Trainer()


class TestHFGeneratorTrainer(HuggingFaceTrainerMixin, BaseGeneratorTrainer):
    __test__ = (
        False  # needed for Pytest collision. Avoids PytestCollectionWarning
    )

    def train(self) -> TrainResult:
        return TrainResult(loss=0.42)

    def evaluate(self) -> TestResult:
        return TestResult(loss=0.42)

    def hf_trainer_obj(self) -> Trainer:
        return Trainer()


@pytest.fixture()
def train_dataset() -> Dataset:
    return Dataset.from_dict(
        {
            "query": ["first query", "second query"],
            "response": ["first response", "second response"],
        }
    )


@pytest.fixture()
def hf_rag_system(mock_rag_system: RAGSystem) -> RAGSystem:
    encoder = SentenceTransformer(modules=[torch.nn.Linear(5, 5)])
    # Mock the tokenize method on the first module
    encoder.tokenizer = None
    encoder._first_module().tokenize = lambda texts: {
        "input_ids": torch.ones((len(texts), 10))
    }
    encoder.encode = lambda texts, **kwargs: torch.ones(
        (len(texts) if isinstance(texts, list) else 1, 5)
    )

    mock_rag_system.retriever.encoder = encoder
    return mock_rag_system



import torch

from fed_rag.utils.data.finetuning_datasets import PyTorchRAGFinetuningDataset


def test_pt_rag_ft_dataset_init(
    input_and_target_ids: tuple[torch.Tensor, torch.Tensor],
) -> None:
    input_ids, target_ids = input_and_target_ids
    rag_ft_dataset = PyTorchRAGFinetuningDataset(
        input_ids=input_ids, target_ids=target_ids
    )

    assert len(rag_ft_dataset) == len(input_ids)
    assert isinstance(rag_ft_dataset, torch.utils.data.Dataset)
    assert rag_ft_dataset[:] == input_and_target_ids[:]



"""PyTorch RAG Finetuning Dataset"""

from typing import Any

import torch
from torch.utils.data import Dataset


class PyTorchRAGFinetuningDataset(Dataset):
    """PyTorch RAG Fine-Tuning Dataset Class.

    Args:
        Dataset (_type_): _description_
    """

    def __init__(
        self, input_ids: list[torch.Tensor], target_ids: list[torch.Tensor]
    ):
        self.input_ids = input_ids
        self.target_ids = target_ids

    def __len__(self) -> int:
        return len(self.input_ids)

    def __getitem__(self, idx: int) -> Any:
        return self.input_ids[idx], self.target_ids[idx]



from contextlib import nullcontext as does_not_raise

import torch

from fed_rag.base.retriever import BaseRetriever


def test_base_abstract_attr() -> None:
    abstract_methods = BaseRetriever.__abstractmethods__

    assert "encode_context" in abstract_methods
    assert "encode_query" in abstract_methods
    assert "encoder" in abstract_methods
    assert "query_encoder" in abstract_methods
    assert "context_encoder" in abstract_methods


def test_base_encode(mock_retriever: BaseRetriever) -> None:
    encoded_ctx = mock_retriever.encode_context("mock context")
    encoded_query = mock_retriever.encode_query("mock query")
    cosine_sim = encoded_ctx @ encoded_query.T
    *_, final_layer = mock_retriever.encoder.parameters()

    with does_not_raise():
        # cosine sim should be a Tensor with a single item
        cosine_sim.item()

    assert encoded_ctx.numel() == final_layer.size()[-1]
    assert encoded_query.numel() == final_layer.size()[-1]
    assert isinstance(mock_retriever.encoder, torch.nn.Module)
    assert mock_retriever.query_encoder is None
    assert mock_retriever.context_encoder is None



"""Tester Decorators"""

from typing import Callable


class TesterDecorators:
    def pytorch(self, func: Callable) -> Callable:
        from fed_rag.inspectors.pytorch import inspect_tester_signature

        def decorator(func: Callable) -> Callable:
            # inspect func sig
            spec = inspect_tester_signature(
                func
            )  # may need to create a cfg for this if decorater accepts params

            # store fl_task config
            func.__setattr__("__fl_task_tester_config", spec)  # type: ignore[attr-defined]

            return func

        return decorator(func)

    def huggingface(self, func: Callable) -> Callable:
        from fed_rag.inspectors.huggingface import inspect_tester_signature

        def decorator(func: Callable) -> Callable:
            # inspect func sig
            spec = inspect_tester_signature(
                func
            )  # may need to create a cfg for this if decorater accepts params

            # store fl_task config
            func.__setattr__("__fl_task_tester_config", spec)  # type: ignore[attr-defined]

            return func

        return decorator(func)



"""HuggingFace Generator Mixin."""

from typing import TYPE_CHECKING, Any, Protocol, Union, runtime_checkable

import torch
import torch.nn.functional as F

if TYPE_CHECKING:  # pragma: no cover
    from peft import PeftModel
    from transformers import PreTrainedModel
    from transformers.generation.utils import GenerationConfig

from fed_rag.tokenizers.hf_pretrained_tokenizer import HFPretrainedTokenizer


@runtime_checkable
class HFGeneratorProtocol(Protocol):
    prompt_template: str
    tokenizer: HFPretrainedTokenizer
    model: Union["PreTrainedModel", "PeftModel"]
    generation_config: "GenerationConfig"


class HuggingFaceGeneratorMixin:
    # generate
    def generate(
        self: HFGeneratorProtocol, query: str, context: str, **kwargs: Any
    ) -> str:
        formatted_query = self.prompt_template.format(
            question=query, context=context
        )

        # encode query
        tokenizer_result = self.tokenizer.unwrapped(
            formatted_query, return_tensors="pt"
        )
        inputs: torch.Tensor = tokenizer_result.input_ids
        inputs = inputs.to(self.model.device)

        # generate
        generated_ids = self.model.generate(
            inputs=inputs,
            generation_config=self.generation_config,
            tokenizer=self.tokenizer.unwrapped,
            **kwargs,
        )

        # decode tokens
        outputs: list[str] = self.tokenizer.unwrapped.batch_decode(
            generated_ids, skip_special_tokens=True
        )
        return outputs[0]

    def compute_target_sequence_proba(
        self: HFGeneratorProtocol, prompt: str, target: str
    ) -> torch.Tensor:
        """Computes the target sequence probability given the prompt.

        Args:
            generator (BaseGenerator): The generator LLM
            prompt (str): The input i.e. conditional prompt sequence
            target (str): The target sequence

        Returns:
            proba (torch.Tensor): The probability of target sequence given a prompt.
                i.e., P_{LLM}(target | prompt)
        """
        input_text = prompt + target
        encode_result = self.tokenizer.encode(input_text)
        input_ids = encode_result["input_ids"]

        # Get the token IDs for just the target portion
        prompt_only_encode_result = self.tokenizer.encode(prompt)
        target_start_idx = len(prompt_only_encode_result["input_ids"])
        target_ids = input_ids[target_start_idx:]

        # Create tensor and send to the device where the model resides
        input_ids_tensor = (
            torch.tensor(input_ids).unsqueeze(0).to(self.model.device)
        )

        # Get the logits from the model
        with torch.no_grad():
            outputs = self.model(input_ids_tensor)
            logits = outputs.logits

        # Calculate probability of each target token given the previous tokens
        log_probs = []
        for i, target_id in enumerate(target_ids):
            # get log prob of next target token in the sequence
            next_token_pos = target_start_idx + i
            next_token_logits = logits[0, next_token_pos, :]
            probs = F.softmax(next_token_logits, dim=-1)
            log_prob = torch.log(probs[target_id]).item()
            log_probs.append(log_prob)

        # Sum log probabilities to get sequence log probability
        sequence_log_prob = sum(log_probs)
        # Convert to probability
        sequence_prob = torch.exp(torch.tensor(sequence_log_prob))

        return sequence_prob



import math

import pytest
import torch

EMB_DIM = 10
BATCH_SIZE = 2
NUM_CHUNKS = 3


@pytest.fixture()
def retrieved_chunks() -> torch.Tensor:
    """Embeddings of 'retrieved' chunks."""
    batch = []
    for bx in range(1, BATCH_SIZE + 1):
        embs = []
        for ix in range(1, NUM_CHUNKS + 1):
            embs.append([bx / ix for _ in range(EMB_DIM)])
        batch.append(embs)

    return torch.tensor(batch, dtype=torch.float32)


@pytest.fixture()
def contexts() -> torch.Tensor:
    batch = []
    for ix in range(1, BATCH_SIZE):
        batch.append(torch.ones(EMB_DIM) * ix)
    return torch.stack(batch, dim=0)


@pytest.fixture()
def lm_scores() -> torch.Tensor:
    """Mock probas of generated outputs 'given' context and chunk."""
    batch = []
    for bx in range(1, BATCH_SIZE + 1):
        scores = [math.exp(ix) for ix in range(NUM_CHUNKS)]
        scores = [el / sum(scores) for el in scores]
        batch.append(scores)

    return torch.tensor(batch, dtype=torch.float32)
